;;; egg.el --- EGG Input Method Architecture

;; Copyright (C) 1997, 1998 Mule Project, Powered by Electrotechnical
;; Laboratory, JAPAN.
;; Project Leader: Satoru Tomura <tomura@etl.go.jp>

;; Author: NIIBE Yutaka <gniibe@mri.co.jp>
;;         KATAYAMA Yoshio <kate@pfu.co.jp>
;; Maintainer: NIIBE Yutaka <gniibe@mri.co.jp>
;; Keywords: mule, multilingual, input method

;; This file will be part of GNU Emacs (in future).

;; EGG is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; EGG is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;; Code:
(defvar egg-mode-preference t
  "Non-nil if modefull.")

(defvar egg-default-language "Japanese")
(defvar egg-last-method-name)
(make-variable-buffer-local 'egg-last-method-name)

;;;###autoload
(defun egg-mode (&rest arg)
  "Toggle EGG  mode.
\\[describe-bindings]
"
  (interactive "P")
  (if (null arg)
      ;; Turn off
      (progn
	(cond
	 ((its-in-fence-p)
	  (its-exit-mode))
	 ((egg-get-bunsetsu-info (point))
	  (egg-exit-conversion)))
	(setq describe-current-input-method-function nil)
	(setq current-input-method nil)
	(use-local-map (keymap-parent (current-local-map)))
	(force-mode-line-update))
    ;; Turn on
    (if (null (string= (car arg) egg-last-method-name))
	(progn
	  (funcall (nth 1 arg))
	  (egg-set-conversion-backend nil (list its-current-language) t)
	  (setq egg-default-language its-current-language)))
    (setq egg-last-method-name (car arg))
    (use-local-map (if egg-mode-preference
		       (egg-modefull-map)
		     (egg-modeless-map)))
    (setq inactivate-current-input-method-function 'egg-mode)
    (setq describe-current-input-method-function 'egg-help)
    (add-hook 'input-method-activate-hook 'its-set-mode-line-title)))

(defun egg-modefull-map ()
  "Generate modefull keymap for EGG mode."  
  (let ((map (make-sparse-keymap))
	(i 33))
    (define-key map "\C-_" 'egg-jis-code-input)
    (while (< i 127)
      (define-key map (vector i) 'egg-self-insert-char)
      (setq i (1+ i)))
    (its-define-select-keys map)
    (set-keymap-parent map (current-local-map))
    map))

(defun egg-modeless-map ()
  "Generate modeless keymap for EGG mode."
  (let ((map (make-sparse-keymap)))
    (define-key map " " 'mlh-space-bar-backward-henkan)
    (define-key map "\C-_" 'egg-jis-code-input)
    (set-keymap-parent map (current-local-map))
    map))

(defun egg-self-insert-char ()
  (interactive)
  (its-start last-command-char))

(defun egg-hinshi-select ()
 (menudiag-select ; Should generate at initialization time
  '(menu  "$BIJ;lL>(B:"
	  (("$BIaDLL>;l(B" .
	    (menu  "$BIJ;lL>(B[$BIaDLL>;l(B]:"
		   ("$BL>;l(B" "$B%59T(B($B$9$k(B)&$BL>;l(B" "$B0lCJ(B&$BL>;l(B"
		    "$B7AMFF0;l(B&$BL>;l(B" "$B?t;l(B")))
	   ("$B8GM-L>;l(B" .
	    (menu  "$BIJ;lL>(B[$B8GM-L>;l(B]:"
		   ("$B?ML>(B" "$BCOL>(B" "$B?ML>(B&$BCOL>(B" "$B8GM-L>;l(B")))
	   ("$BF0;l(B" .
	    (menu  "$BIJ;lL>(B[$BF0;l(B]:"
		   ("$B0lCJ(B" "$B0lCJ(B&$BL>;l(B" "$B%+9T8^CJ(B" "$B%,9T8^CJ(B"
		    "$B%59T8^CJ(B" "$B%?9T8^CJ(B")))
	   ("$BFC<l$JF0;l(B" .
	    (menu  "$BIJ;lL>(B[$BFC<l$JF0;l(B]:"
		   ("$B%+9T(B($B9T$/(B)" "$B%i9T(B($B2<$5$$(B)" "$BMh(B($B$3(B)"
		    "$BMh(B($B$-(B)" "$BMh(B($B$/(B)" "$B0Y(B($B$7(B)")))
	   ("$BF0;l0J30$NMQ8@(B" .
	    (menu  "$BIJ;lL>(B[$BF0;l0J30$NMQ8@(B]:"
		   ("$B7AMF;l(B" "$B7AMFF0;l(B" "$B7AMFF0;l(B&$BL>;l(B"
		    "$B7AMFF0;l(B($B$?$k(B)")))))))

;; XXX: Should use backend interface
(defun egg-toroku-region (start end)
  (interactive "r")
  (let* ((env (wnn-get-environment wnn-dictionary-specification)) ; XXX
	 (kanji (buffer-substring start end))
	 (yomi (read-multilingual-string 
		(format "$B<-=qEPO?!X(B%s$B!Y(B  $BFI$_(B:" kanji)))
	 (dic (menudiag-select (list 'menu "$BEPO?<-=qL>(B:"
				 ;; XXX
				 (wnn-list-writable-dictionaries-byname env))))
	 (dic-name (wnn-dict-name dic))
	 (hinshi (egg-hinshi-select))
	 (hinshi-id (wnn-hinshi-number env hinshi)))
    (if (y-or-n-p
	 (format "$B<-=q9`L\!X(B%s$B!Y(B(%s: %s)$B$r(B %s $B$KEPO?$7$^$9(B"
		 kanji yomi hinshi dic-name))
	(let ((r (wnn-add-word env dic yomi kanji "" hinshi-id 0)))
	  (if (< r 0)
	    (error "$B<-=qEPO?!X(B%s$B!Y(B(%s: %s) %s $B$K<:GT$7$^$7$?(B: %s"
		   kanji yomi hinshi dic-name
		   (wnnrpc-get-error-message (- r)))
	    (message 
	     "$B<-=q9`L\!X(B%s$B!Y(B(%s: %s)$B$r(B %s $B$KEPO?$7$^$7$?(B" 
	     kanji yomi hinshi dic-name))))))

;;;
;;; auto fill controll
;;;

(defun egg-do-auto-fill ()
  (if (and auto-fill-function (> (current-column) fill-column))
      (let ((ocolumn (current-column)))
	(funcall auto-fill-function)
	(while (and (< fill-column (current-column))
		    (< (current-column) ocolumn))
  	  (setq ocolumn (current-column))
	  (funcall auto-fill-function)))))

(require 'its)
(require 'menudiag)
(require 'egg-mlh)
(require 'egg-cnv)
(require 'egg-com)
(require 'custom)

(defgroup egg nil
  "Tamagotchy --- EGG Versio 4.0")

;;(load-library "its/hira")
;;(setq-default its-current-map its-hira-map)

;;(load-library "egg/wnn")
;;(load-library "egg/wnnrpc")
;;(setq egg-conversion-backend wnn-conversion-backend)

;;(load-library "egg/sj3rpc")
;;(load-library "egg/sj3")
;;(setq egg-conversion-backend sj3-conversion-backend)

(defvar egg-support-languages nil)

(defun egg-set-support-languages (langs)
  (while langs
    (if (null (member (car langs) egg-support-languages))
	(setq egg-support-languages (cons (car langs) egg-support-languages)))
    (setq langs (cdr langs))))

(add-hook 'kill-emacs-hook 'egg-kill-emacs-function)
(defun egg-kill-emacs-function ()
  (egg-finalize-backend))

(provide 'egg)

;;; egg.el ends here
