;;; its/hangul.el --- Hangul Input in Egg Input Method Architecture

;; Copyright (C) 1997, 1998 Mule Project,
;; Powered by Electrotechnical Laboratory, JAPAN.
;; Project Leader: Satoru Tomura <tomura@etl.go.jp>

;; Author: NIIBE Yutaka <gniibe@mri.co.jp>
;;         KATAYAMA Yoshio <kate@pfu.co.jp>

;; This file will be part of GNU Emacs (in future).

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:


;;; Code:

(eval-when-compile
  (require 'its)
  (require 'cl))

(defvar its-hangul-enable-jeonkak-alphabet t "*Enable Jeonkak alphabet")
(defvar	its-hangul-period ". " "*$(C#.(B")	; ". " "$(C#.(B"
(defvar	its-hangul-comma  ", " "*$(C#,(B")	; ", " "$(C#,(B"
(defvar its-hangul-open-bracket  "$(C!8(B" "*[")  ; "$(C#[(B"
(defvar its-hangul-close-bracket "$(C!9(B" "*]")  ; "$(C#](B"

(eval-when-compile
 (defconst its-hangul-kijuhynbmlopOP
   (its-make-class+back "[kijuhynbmlopOP]" -2))
 (defconst its-hangul-kijuhynbmlopP
   (its-make-class+back "[kijuhynbmlopP]" -2))
 (defconst its-hangul-kjhynmlop
   (its-make-class+back "[kjhynmlop]"	 -2))

 (defmacro its-defrule-hangul-2 (seq out &rest exprs)
   `(let ((next-state (its-goto-state ,seq nil t)))
      (its-set-output next-state ,out)
      (its-make-next-state next-state -1 ,seq ,out)
      (its-defrule-otherwise next-state ,out)
      (let ((kijuhynbmlopOP (its-make-otherwise ,out its-hangul-kijuhynbmlopOP))
	    (kijuhynbmlopP  (its-make-otherwise ,out its-hangul-kijuhynbmlopP))
	    (kjhynmlop      (its-make-otherwise ,out its-hangul-kjhynmlop))
	    (next-state next-state))
	,@exprs)))

 (defmacro its-define-hangul (seq out otherwise &rest exprs)
   `(let ((next-state (its-make-next-state ',next-state
					   (aref ,seq (1- (length ,seq)))
					   ,seq ,out)))
      (its-set-output next-state ,out)
      (its-make-next-state next-state -1 ,seq ,out)
      (its-define-otherwise next-state
			    (its-make-otherwise ,out its-otherwise-back-one))
      (its-define-otherwise next-state ,otherwise)
      (if ',exprs
	  (let ((kijuhynbmlopOP (its-make-otherwise ,out its-hangul-kijuhynbmlopOP))
		(kijuhynbmlopP  (its-make-otherwise ,out its-hangul-kijuhynbmlopP))
		(kjhynmlop      (its-make-otherwise ,out its-hangul-kjhynmlop)))
	    ,@exprs))))

 (defun its-defrule-hangul (seq output)
   (let ((state (its-goto-state seq nil t)))
     (its-set-output state output)
     (its-make-next-state state -1 seq output)

     (its-defrule-otherwise state output))))

(define-its-state-machine its-hangul-map
  "hangul" "$(CGQ(B" "Korean"
  "Map for Hangul input. (Korean)"

  (defconst its-jeonkak-escape "Z")
  (defconst its-pankak-escape  "X")

  (its-defrule-select-mode-temporally "B" downcase)
  (its-defrule-select-mode-temporally "J" jeonkak-downcase)

  (its-defrule-hangul	"r"	"$(C$!(B")
  (its-defrule-hangul	"s"	"$(C$$(B")
  (its-defrule-hangul	"e"	"$(C$'(B")
  (its-defrule-hangul	"f"	"$(C$)(B")
  (its-defrule-hangul	"a"	"$(C$1(B")
  (its-defrule-hangul	"q"	"$(C$2(B")
  (its-defrule-hangul	"t"	"$(C$5(B")
  (its-defrule-hangul	"d"	"$(C$7(B")
  (its-defrule-hangul	"w"	"$(C$8(B")
  (its-defrule-hangul	"c"	"$(C$:(B")
  (its-defrule-hangul	"z"	"$(C$;(B")
  (its-defrule-hangul	"x"	"$(C$<(B")
  (its-defrule-hangul	"v"	"$(C$=(B")
  (its-defrule-hangul	"g"	"$(C$>(B")
  (its-defrule-hangul	"R"	"$(C$"(B")
  (its-defrule-hangul	"E"	"$(C$((B")
  (its-defrule-hangul	"Q"	"$(C$3(B")
  (its-defrule-hangul	"T"	"$(C$6(B")
  (its-defrule-hangul	"W"	"$(C$9(B")

  (its-defrule-hangul	"k"	"$(C$?(B")
  (its-defrule-hangul	"i"	"$(C$A(B")
  (its-defrule-hangul	"j"	"$(C$C(B")
  (its-defrule-hangul	"u"	"$(C$E(B")
  (its-defrule-hangul	"h"	"$(C$G(B")
  (its-defrule-hangul	"y"	"$(C$K(B")
  (its-defrule-hangul	"n"	"$(C$L(B")
  (its-defrule-hangul	"b"	"$(C$P(B")
  (its-defrule-hangul	"m"	"$(C$Q(B")
  (its-defrule-hangul	"l"	"$(C$S(B")
  (its-defrule-hangul	"o"	"$(C$@(B")
  (its-defrule-hangul	"p"	"$(C$D(B")
  (its-defrule-hangul	"O"	"$(C$B(B")
  (its-defrule-hangul	"P"	"$(C$F(B")

  (its-defrule-hangul-2	"rk"	"$(C0!(B"
    (its-define-hangul	"rkr"	"$(C0"(B"	kijuhynbmlopOP)
    (its-define-hangul	"rks"	"$(C0#(B"	kijuhynbmlopP)
    (its-define-hangul	"rke"	"$(C0$(B"	kijuhynbmlopP)
    (its-define-hangul	"rkf"	"$(C0%(B"	kijuhynbmlopP
     (its-define-hangul	"rkfr"	"$(C0&(B"	kijuhynbmlopOP)
     (its-define-hangul	"rkfa"	"$(C0'(B"	kijuhynbmlopP))
    (its-define-hangul	"rka"	"$(C0((B"	kijuhynbmlopP)
    (its-define-hangul	"rkq"	"$(C0)(B"	kijuhynbmlopP
     (its-define-hangul	"rkqt"	"$(C0*(B"	kijuhynbmlopOP))
    (its-define-hangul	"rkt"	"$(C0+(B"	kijuhynbmlopOP)
    (its-define-hangul	"rkT"	"$(C0,(B"	kjhynmlop)
    (its-define-hangul	"rkd"	"$(C0-(B"	kijuhynbmlopOP)
    (its-define-hangul	"rkw"	"$(C0.(B"	kijuhynbmlopOP)
    (its-define-hangul	"rkc"	"$(C0/(B"	kijuhynbmlopP)
    (its-define-hangul	"rkx"	"$(C00(B"	kijuhynbmlopP)
    (its-define-hangul	"rkv"	"$(C01(B"	kijuhynbmlopP)
    (its-define-hangul	"rkg"	"$(C02(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ro"	"$(C03(B"
    (its-define-hangul	"ror"	"$(C04(B"	kijuhynbmlopOP)
    (its-define-hangul	"ros"	"$(C05(B"	kijuhynbmlopP)
    (its-define-hangul	"rof"	"$(C06(B"	kijuhynbmlopP)
    (its-define-hangul	"roa"	"$(C07(B"	kijuhynbmlopP)
    (its-define-hangul	"roq"	"$(C08(B"	kijuhynbmlopP)
    (its-define-hangul	"rot"	"$(C09(B"	kijuhynbmlopOP)
    (its-define-hangul	"roT"	"$(C0:(B"	kjhynmlop)
    (its-define-hangul	"rod"	"$(C0;(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ri"	"$(C0<(B"
    (its-define-hangul	"rir"	"$(C0=(B"	kijuhynbmlopOP)
    (its-define-hangul	"ris"	"$(C0>(B"	kijuhynbmlopP)
    (its-define-hangul	"rif"	"$(C0?(B"	kijuhynbmlopP)
    (its-define-hangul	"rit"	"$(C0@(B"	kijuhynbmlopOP)
    (its-define-hangul	"rid"	"$(C0A(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rO"	"$(C0B(B"
    (its-define-hangul	"rOs"	"$(C0C(B"	kijuhynbmlopP)
    (its-define-hangul	"rOf"	"$(C0D(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"rj"	"$(C0E(B"
    (its-define-hangul	"rjr"	"$(C0F(B"	kijuhynbmlopOP)
    (its-define-hangul	"rjs"	"$(C0G(B"	kijuhynbmlopP)
    (its-define-hangul	"rje"	"$(C0H(B"	kijuhynbmlopP)
    (its-define-hangul	"rjf"	"$(C0I(B"	kijuhynbmlopP
     (its-define-hangul	"rjfa"	"$(C0J(B"	kijuhynbmlopP))
    (its-define-hangul	"rja"	"$(C0K(B"	kijuhynbmlopP)
    (its-define-hangul	"rjq"	"$(C0L(B"	kijuhynbmlopP)
    (its-define-hangul	"rjt"	"$(C0M(B"	kijuhynbmlopOP)
    (its-define-hangul	"rjT"	"$(C0N(B"	kjhynmlop)
    (its-define-hangul	"rjd"	"$(C0O(B"	kijuhynbmlopOP)
    (its-define-hangul	"rjw"	"$(C0P(B"	kijuhynbmlopOP)
    (its-define-hangul	"rjx"	"$(C0Q(B"	kijuhynbmlopP)
    (its-define-hangul	"rjv"	"$(C0R(B"	kijuhynbmlopP)
    (its-define-hangul	"rjg"	"$(C0S(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"rp"	"$(C0T(B"
    (its-define-hangul	"rps"	"$(C0U(B"	kijuhynbmlopP)
    (its-define-hangul	"rpf"	"$(C0V(B"	kijuhynbmlopP)
    (its-define-hangul	"rpa"	"$(C0W(B"	kijuhynbmlopP)
    (its-define-hangul	"rpq"	"$(C0X(B"	kijuhynbmlopP)
    (its-define-hangul	"rpt"	"$(C0Y(B"	kijuhynbmlopOP)
    (its-define-hangul	"rpT"	"$(C0Z(B"	kjhynmlop)
    (its-define-hangul	"rpd"	"$(C0[(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ru"	"$(C0\(B"
    (its-define-hangul	"rur"	"$(C0](B"	kijuhynbmlopOP)
    (its-define-hangul	"ruR"	"$(C0^(B"	kijuhynbmlopP)
    (its-define-hangul	"rus"	"$(C0_(B"	kijuhynbmlopP)
    (its-define-hangul	"rue"	"$(C0`(B"	kijuhynbmlopP)
    (its-define-hangul	"ruf"	"$(C0a(B"	kijuhynbmlopP)
    (its-define-hangul	"rua"	"$(C0b(B"	kijuhynbmlopP)
    (its-define-hangul	"ruq"	"$(C0c(B"	kijuhynbmlopP)
    (its-define-hangul	"rut"	"$(C0d(B"	kijuhynbmlopOP)
    (its-define-hangul	"ruT"	"$(C0e(B"	kjhynmlop)
    (its-define-hangul	"rud"	"$(C0f(B"	kijuhynbmlopOP)
    (its-define-hangul	"rux"	"$(C0g(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"rP"	"$(C0h(B"
    (its-define-hangul	"rPs"	"$(C0i(B"	kijuhynbmlopP)
    (its-define-hangul	"rPf"	"$(C0j(B"	kijuhynbmlopP)
    (its-define-hangul	"rPq"	"$(C0k(B"	kijuhynbmlopP)
    (its-define-hangul	"rPt"	"$(C0l(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rh"	"$(C0m(B"
    (its-define-hangul	"rhr"	"$(C0n(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhs"	"$(C0o(B"	kijuhynbmlopP)
    (its-define-hangul	"rhe"	"$(C0p(B"	kijuhynbmlopP)
    (its-define-hangul	"rhf"	"$(C0q(B"	kijuhynbmlopP
     (its-define-hangul	"rhfa"	"$(C0r(B"	kijuhynbmlopP)
     (its-define-hangul	"rhft"	"$(C0s(B"	kijuhynbmlopOP)
     (its-define-hangul	"rhfg"	"$(C0t(B"	kijuhynbmlopP))
    (its-define-hangul	"rha"	"$(C0u(B"	kijuhynbmlopP)
    (its-define-hangul	"rhq"	"$(C0v(B"	kijuhynbmlopP)
    (its-define-hangul	"rht"	"$(C0w(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhd"	"$(C0x(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhw"	"$(C0y(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rhk"	"$(C0z(B"
    (its-define-hangul	"rhkr"	"$(C0{(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhks"	"$(C0|(B"	kijuhynbmlopP)
    (its-define-hangul	"rhkf"	"$(C0}(B"	kijuhynbmlopP
     (its-define-hangul	"rhkfa"	"$(C0~(B"	kijuhynbmlopP))
    (its-define-hangul	"rhka"	"$(C1!(B"	kijuhynbmlopP)
    (its-define-hangul	"rhkq"	"$(C1"(B"	kijuhynbmlopP)
    (its-define-hangul	"rhkt"	"$(C1#(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhkd"	"$(C1$(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rho"	"$(C1%(B"
    (its-define-hangul	"rhos"	"$(C1&(B"	kijuhynbmlopP)
    (its-define-hangul	"rhof"	"$(C1'(B"	kijuhynbmlopP)
    (its-define-hangul	"rhoq"	"$(C1((B"	kijuhynbmlopP)
    (its-define-hangul	"rhoT"	"$(C1)(B"	kjhynmlop)
    (its-define-hangul	"rhod"	"$(C1*(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rhl"	"$(C1+(B"
    (its-define-hangul	"rhlr"	"$(C1,(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhls"	"$(C1-(B"	kijuhynbmlopP)
    (its-define-hangul	"rhlf"	"$(C1.(B"	kijuhynbmlopP)
    (its-define-hangul	"rhla"	"$(C1/(B"	kijuhynbmlopP)
    (its-define-hangul	"rhlq"	"$(C10(B"	kijuhynbmlopP)
    (its-define-hangul	"rhlt"	"$(C11(B"	kijuhynbmlopOP)
    (its-define-hangul	"rhld"	"$(C12(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ry"	"$(C13(B"
    (its-define-hangul	"rys"	"$(C14(B"	kijuhynbmlopP)
    (its-define-hangul	"ryf"	"$(C15(B"	kijuhynbmlopP)
    (its-define-hangul	"ryq"	"$(C16(B"	kijuhynbmlopP)
    (its-define-hangul	"ryt"	"$(C17(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rn"	"$(C18(B"
    (its-define-hangul	"rnr"	"$(C19(B"	kijuhynbmlopOP)
    (its-define-hangul	"rns"	"$(C1:(B"	kijuhynbmlopP)
    (its-define-hangul	"rne"	"$(C1;(B"	kijuhynbmlopP)
    (its-define-hangul	"rnf"	"$(C1<(B"	kijuhynbmlopP
     (its-define-hangul	"rnfr"	"$(C1=(B"	kijuhynbmlopOP)
     (its-define-hangul	"rnfa"	"$(C1>(B"	kijuhynbmlopP)
     (its-define-hangul	"rnfg"	"$(C1?(B"	kijuhynbmlopP))
    (its-define-hangul	"rna"	"$(C1@(B"	kijuhynbmlopP)
    (its-define-hangul	"rnq"	"$(C1A(B"	kijuhynbmlopP)
    (its-define-hangul	"rnt"	"$(C1B(B"	kijuhynbmlopOP)
    (its-define-hangul	"rnd"	"$(C1C(B"	kijuhynbmlopOP)
    (its-define-hangul	"rnw"	"$(C1D(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rnj"	"$(C1E(B"
    (its-define-hangul	"rnjr"	"$(C1F(B"	kijuhynbmlopOP)
    (its-define-hangul	"rnjs"	"$(C1G(B"	kijuhynbmlopP)
    (its-define-hangul	"rnjf"	"$(C1H(B"	kijuhynbmlopP)
    (its-define-hangul	"rnjT"	"$(C1I(B"	kjhynmlop)
    (its-define-hangul	"rnjd"	"$(C1J(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rnp"	"$(C1K(B"
    (its-define-hangul	"rnpt"	"$(C1L(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rnl"	"$(C1M(B"
    (its-define-hangul	"rnlr"	"$(C1N(B"	kijuhynbmlopOP)
    (its-define-hangul	"rnls"	"$(C1O(B"	kijuhynbmlopP)
    (its-define-hangul	"rnlf"	"$(C1P(B"	kijuhynbmlopP)
    (its-define-hangul	"rnla"	"$(C1Q(B"	kijuhynbmlopP)
    (its-define-hangul	"rnlq"	"$(C1R(B"	kijuhynbmlopP)
    (its-define-hangul	"rnlt"	"$(C1S(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"rb"	"$(C1T(B"
    (its-define-hangul	"rbs"	"$(C1U(B"	kijuhynbmlopP)
    (its-define-hangul	"rbf"	"$(C1V(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"rm"	"$(C1W(B"
    (its-define-hangul	"rmr"	"$(C1X(B"	kijuhynbmlopOP)
    (its-define-hangul	"rms"	"$(C1Y(B"	kijuhynbmlopP)
    (its-define-hangul	"rme"	"$(C1Z(B"	kijuhynbmlopP)
    (its-define-hangul	"rmf"	"$(C1[(B"	kijuhynbmlopP
     (its-define-hangul	"rmfr"	"$(C1\(B"	kijuhynbmlopOP))
    (its-define-hangul	"rma"	"$(C1](B"	kijuhynbmlopP)
    (its-define-hangul	"rmq"	"$(C1^(B"	kijuhynbmlopP)
    (its-define-hangul	"rmt"	"$(C1_(B"	kijuhynbmlopOP)
    (its-define-hangul	"rmd"	"$(C1`(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"rml"	"$(C1a(B")

  (its-defrule-hangul-2	"rl"	"$(C1b(B"
    (its-define-hangul	"rlr"	"$(C1c(B"	kijuhynbmlopOP)
    (its-define-hangul	"rls"	"$(C1d(B"	kijuhynbmlopP)
    (its-define-hangul	"rle"	"$(C1e(B"	kijuhynbmlopP)
    (its-define-hangul	"rlf"	"$(C1f(B"	kijuhynbmlopP
     (its-define-hangul	"rlfa"	"$(C1g(B"	kijuhynbmlopP))
    (its-define-hangul	"rla"	"$(C1h(B"	kijuhynbmlopP)
    (its-define-hangul	"rlq"	"$(C1i(B"	kijuhynbmlopP)
    (its-define-hangul	"rlt"	"$(C1j(B"	kijuhynbmlopOP)
    (its-define-hangul	"rld"	"$(C1k(B"	kijuhynbmlopOP)
    (its-define-hangul	"rlw"	"$(C1l(B"	kijuhynbmlopOP)
    (its-define-hangul	"rlv"	"$(C1m(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Rk"	"$(C1n(B"
    (its-define-hangul	"Rkr"	"$(C1o(B"	kijuhynbmlopOP)
    (its-define-hangul	"RkR"	"$(C1p(B"	kijuhynbmlopP)
    (its-define-hangul	"Rks"	"$(C1q(B"	kijuhynbmlopP)
    (its-define-hangul	"Rkf"	"$(C1r(B"	kijuhynbmlopP
     (its-define-hangul	"Rkfa"	"$(C1s(B"	kijuhynbmlopP))
    (its-define-hangul	"Rka"	"$(C1t(B"	kijuhynbmlopP)
    (its-define-hangul	"Rkq"	"$(C1u(B"	kijuhynbmlopP)
    (its-define-hangul	"Rkt"	"$(C1v(B"	kijuhynbmlopOP)
    (its-define-hangul	"RkT"	"$(C1w(B"	kjhynmlop)
    (its-define-hangul	"Rkd"	"$(C1x(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rkx"	"$(C1y(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Ro"	"$(C1z(B"
    (its-define-hangul	"Ror"	"$(C1{(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ros"	"$(C1|(B"	kijuhynbmlopP)
    (its-define-hangul	"Rof"	"$(C1}(B"	kijuhynbmlopP)
    (its-define-hangul	"Roa"	"$(C1~(B"	kijuhynbmlopP)
    (its-define-hangul	"Roq"	"$(C2!(B"	kijuhynbmlopP)
    (its-define-hangul	"Rot"	"$(C2"(B"	kijuhynbmlopOP)
    (its-define-hangul	"RoT"	"$(C2#(B"	kjhynmlop)
    (its-define-hangul	"Rod"	"$(C2$(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Ri"	"$(C2%(B"
    (its-define-hangul	"Rir"	"$(C2&(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rif"	"$(C2'(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Rj"	"$(C2((B"
    (its-define-hangul	"Rjr"	"$(C2)(B"	kijuhynbmlopOP)
    (its-define-hangul	"RjR"	"$(C2*(B"	kijuhynbmlopP)
    (its-define-hangul	"Rjs"	"$(C2+(B"	kijuhynbmlopP)
    (its-define-hangul	"Rjf"	"$(C2,(B"	kijuhynbmlopP)
    (its-define-hangul	"Rja"	"$(C2-(B"	kijuhynbmlopP)
    (its-define-hangul	"Rjq"	"$(C2.(B"	kijuhynbmlopP)
    (its-define-hangul	"Rjt"	"$(C2/(B"	kijuhynbmlopOP)
    (its-define-hangul	"RjT"	"$(C20(B"	kjhynmlop)
    (its-define-hangul	"Rjd"	"$(C21(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Rp"	"$(C22(B"
    (its-define-hangul	"Rpr"	"$(C23(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rps"	"$(C24(B"	kijuhynbmlopP)
    (its-define-hangul	"Rpa"	"$(C25(B"	kijuhynbmlopP)
    (its-define-hangul	"Rpt"	"$(C26(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rpd"	"$(C27(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Ru"	"$(C28(B"
    (its-define-hangul	"Rus"	"$(C29(B"	kijuhynbmlopP)
    (its-define-hangul	"Ruf"	"$(C2:(B"	kijuhynbmlopP)
    (its-define-hangul	"Rut"	"$(C2;(B"	kijuhynbmlopOP)
    (its-define-hangul	"RuT"	"$(C2<(B"	kjhynmlop)
    (its-define-hangul	"Rux"	"$(C2=(B"	kijuhynbmlopP))

  (its-defrule-hangul	"RP"	"$(C2>(B")

  (its-defrule-hangul-2	"Rh"	"$(C2?(B"
    (its-define-hangul	"Rhr"	"$(C2@(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhs"	"$(C2A(B"	kijuhynbmlopP
     (its-define-hangul	"Rhsg"	"$(C2B(B"	kijuhynbmlopP))
    (its-define-hangul	"Rhf"	"$(C2C(B"	kijuhynbmlopP)
    (its-define-hangul	"Rha"	"$(C2D(B"	kijuhynbmlopP)
    (its-define-hangul	"Rhq"	"$(C2E(B"	kijuhynbmlopP)
    (its-define-hangul	"Rht"	"$(C2F(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhd"	"$(C2G(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhw"	"$(C2H(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhc"	"$(C2I(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Rhk"	"$(C2J(B"
    (its-define-hangul	"Rhkr"	"$(C2K(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhkf"	"$(C2L(B"	kijuhynbmlopP)
    (its-define-hangul	"RhkT"	"$(C2M(B"	kjhynmlop)
    (its-define-hangul	"Rhkd"	"$(C2N(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Rho"	"$(C2O(B"
    (its-define-hangul	"Rhor"	"$(C2P(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rhod"	"$(C2Q(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Rhl"	"$(C2R(B"
    (its-define-hangul	"Rhls"	"$(C2S(B"	kijuhynbmlopP)
    (its-define-hangul	"Rhlf"	"$(C2T(B"	kijuhynbmlopP)
    (its-define-hangul	"Rhla"	"$(C2U(B"	kijuhynbmlopP)
    (its-define-hangul	"Rhlq"	"$(C2V(B"	kijuhynbmlopP)
    (its-define-hangul	"Rhld"	"$(C2W(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"Ry"	"$(C2X(B")

  (its-defrule-hangul-2	"Rn"	"$(C2Y(B"
    (its-define-hangul	"Rnr"	"$(C2Z(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rns"	"$(C2[(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnf"	"$(C2\(B"	kijuhynbmlopP
     (its-define-hangul	"Rnfg"	"$(C2](B"	kijuhynbmlopP))
    (its-define-hangul	"Rna"	"$(C2^(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnq"	"$(C2_(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnt"	"$(C2`(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rnd"	"$(C2a(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rnw"	"$(C2b(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Rnj"	"$(C2c(B"
    (its-define-hangul	"Rnjf"	"$(C2d(B"	kijuhynbmlopP)
    (its-define-hangul	"RnjT"	"$(C2e(B"	kjhynmlop)
    (its-define-hangul	"Rnjd"	"$(C2f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Rnp"	"$(C2g(B"
    (its-define-hangul	"Rnpr"	"$(C2h(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rnps"	"$(C2i(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnpf"	"$(C2j(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnpa"	"$(C2k(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnpq"	"$(C2l(B"	kijuhynbmlopP)
    (its-define-hangul	"RnpT"	"$(C2m(B"	kjhynmlop))

  (its-defrule-hangul-2	"Rnl"	"$(C2n(B"
    (its-define-hangul	"Rnls"	"$(C2o(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnlf"	"$(C2p(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnla"	"$(C2q(B"	kijuhynbmlopP)
    (its-define-hangul	"Rnlq"	"$(C2r(B"	kijuhynbmlopP))

  (its-defrule-hangul	"Rb"	"$(C2s(B")

  (its-defrule-hangul-2	"Rm"	"$(C2t(B"
    (its-define-hangul	"Rmr"	"$(C2u(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rms"	"$(C2v(B"	kijuhynbmlopP
     (its-define-hangul	"Rmsg"	"$(C2w(B"	kijuhynbmlopP))
    (its-define-hangul	"Rmf"	"$(C2x(B"	kijuhynbmlopP
     (its-define-hangul	"Rmfa"	"$(C2y(B"	kijuhynbmlopP)
     (its-define-hangul	"Rmfg"	"$(C2z(B"	kijuhynbmlopP))
    (its-define-hangul	"Rma"	"$(C2{(B"	kijuhynbmlopP)
    (its-define-hangul	"Rmq"	"$(C2|(B"	kijuhynbmlopP)
    (its-define-hangul	"Rmt"	"$(C2}(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rmd"	"$(C2~(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rmx"	"$(C3!(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Rl"	"$(C3"(B"
    (its-define-hangul	"Rlr"	"$(C3#(B"	kijuhynbmlopOP)
    (its-define-hangul	"Rls"	"$(C3$(B"	kijuhynbmlopP)
    (its-define-hangul	"Rlf"	"$(C3%(B"	kijuhynbmlopP)
    (its-define-hangul	"Rla"	"$(C3&(B"	kijuhynbmlopP)
    (its-define-hangul	"Rlq"	"$(C3'(B"	kijuhynbmlopP)
    (its-define-hangul	"Rlt"	"$(C3((B"	kijuhynbmlopOP)
    (its-define-hangul	"Rld"	"$(C3)(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"sk"	"$(C3*(B"
    (its-define-hangul	"skr"	"$(C3+(B"	kijuhynbmlopOP)
    (its-define-hangul	"skR"	"$(C3,(B"	kijuhynbmlopP)
    (its-define-hangul	"sks"	"$(C3-(B"	kijuhynbmlopP)
    (its-define-hangul	"ske"	"$(C3.(B"	kijuhynbmlopP)
    (its-define-hangul	"skf"	"$(C3/(B"	kijuhynbmlopP
     (its-define-hangul	"skfr"	"$(C30(B"	kijuhynbmlopOP)
     (its-define-hangul	"skfa"	"$(C31(B"	kijuhynbmlopP))
    (its-define-hangul	"ska"	"$(C32(B"	kijuhynbmlopP)
    (its-define-hangul	"skq"	"$(C33(B"	kijuhynbmlopP)
    (its-define-hangul	"skt"	"$(C34(B"	kijuhynbmlopOP)
    (its-define-hangul	"skT"	"$(C35(B"	kjhynmlop)
    (its-define-hangul	"skd"	"$(C36(B"	kijuhynbmlopOP)
    (its-define-hangul	"skw"	"$(C37(B"	kijuhynbmlopOP)
    (its-define-hangul	"skc"	"$(C38(B"	kijuhynbmlopP)
    (its-define-hangul	"skx"	"$(C39(B"	kijuhynbmlopP)
    (its-define-hangul	"skg"	"$(C3:(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"so"	"$(C3;(B"
    (its-define-hangul	"sor"	"$(C3<(B"	kijuhynbmlopOP)
    (its-define-hangul	"sos"	"$(C3=(B"	kijuhynbmlopP)
    (its-define-hangul	"sof"	"$(C3>(B"	kijuhynbmlopP)
    (its-define-hangul	"soa"	"$(C3?(B"	kijuhynbmlopP)
    (its-define-hangul	"soq"	"$(C3@(B"	kijuhynbmlopP)
    (its-define-hangul	"sot"	"$(C3A(B"	kijuhynbmlopOP)
    (its-define-hangul	"soT"	"$(C3B(B"	kjhynmlop)
    (its-define-hangul	"sod"	"$(C3C(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"si"	"$(C3D(B"
    (its-define-hangul	"sir"	"$(C3E(B"	kijuhynbmlopOP)
    (its-define-hangul	"sis"	"$(C3F(B"	kijuhynbmlopP)
    (its-define-hangul	"sif"	"$(C3G(B"	kijuhynbmlopP)
    (its-define-hangul	"sia"	"$(C3H(B"	kijuhynbmlopP)
    (its-define-hangul	"sid"	"$(C3I(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"sj"	"$(C3J(B"
    (its-define-hangul	"sjr"	"$(C3K(B"	kijuhynbmlopOP
     (its-define-hangul	"sjrt"	"$(C3L(B"	kijuhynbmlopOP))
    (its-define-hangul	"sjs"	"$(C3M(B"	kijuhynbmlopP)
    (its-define-hangul	"sjf"	"$(C3N(B"	kijuhynbmlopP
     (its-define-hangul	"sjfa"	"$(C3O(B"	kijuhynbmlopP)
     (its-define-hangul	"sjfq"	"$(C3P(B"	kijuhynbmlopP))
    (its-define-hangul	"sja"	"$(C3Q(B"	kijuhynbmlopP)
    (its-define-hangul	"sjq"	"$(C3R(B"	kijuhynbmlopP)
    (its-define-hangul	"sjt"	"$(C3S(B"	kijuhynbmlopOP)
    (its-define-hangul	"sjT"	"$(C3T(B"	kjhynmlop)
    (its-define-hangul	"sjd"	"$(C3U(B"	kijuhynbmlopOP)
    (its-define-hangul	"sjg"	"$(C3V(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sp"	"$(C3W(B"
    (its-define-hangul	"spr"	"$(C3X(B"	kijuhynbmlopOP)
    (its-define-hangul	"sps"	"$(C3Y(B"	kijuhynbmlopP)
    (its-define-hangul	"spf"	"$(C3Z(B"	kijuhynbmlopP)
    (its-define-hangul	"spa"	"$(C3[(B"	kijuhynbmlopP)
    (its-define-hangul	"spq"	"$(C3\(B"	kijuhynbmlopP)
    (its-define-hangul	"spt"	"$(C3](B"	kijuhynbmlopOP)
    (its-define-hangul	"spT"	"$(C3^(B"	kjhynmlop)
    (its-define-hangul	"spd"	"$(C3_(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"su"	"$(C3`(B"
    (its-define-hangul	"sur"	"$(C3a(B"	kijuhynbmlopOP)
    (its-define-hangul	"sus"	"$(C3b(B"	kijuhynbmlopP)
    (its-define-hangul	"suf"	"$(C3c(B"	kijuhynbmlopP)
    (its-define-hangul	"sua"	"$(C3d(B"	kijuhynbmlopP)
    (its-define-hangul	"suq"	"$(C3e(B"	kijuhynbmlopP)
    (its-define-hangul	"suT"	"$(C3f(B"	kjhynmlop)
    (its-define-hangul	"sud"	"$(C3g(B"	kijuhynbmlopOP)
    (its-define-hangul	"suz"	"$(C3h(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sP"	"$(C3i(B"
    (its-define-hangul	"sPs"	"$(C3j(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sh"	"$(C3k(B"
    (its-define-hangul	"shr"	"$(C3l(B"	kijuhynbmlopOP)
    (its-define-hangul	"shs"	"$(C3m(B"	kijuhynbmlopP)
    (its-define-hangul	"shf"	"$(C3n(B"	kijuhynbmlopP
     (its-define-hangul	"shfa"	"$(C3o(B"	kijuhynbmlopP))
    (its-define-hangul	"sha"	"$(C3p(B"	kijuhynbmlopP)
    (its-define-hangul	"shq"	"$(C3q(B"	kijuhynbmlopP)
    (its-define-hangul	"sht"	"$(C3r(B"	kijuhynbmlopOP)
    (its-define-hangul	"shd"	"$(C3s(B"	kijuhynbmlopOP)
    (its-define-hangul	"shv"	"$(C3t(B"	kijuhynbmlopP)
    (its-define-hangul	"shg"	"$(C3u(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"shk"	"$(C3v(B"
    (its-define-hangul	"shks"	"$(C3w(B"	kijuhynbmlopP)
    (its-define-hangul	"shkf"	"$(C3x(B"	kijuhynbmlopP)
    (its-define-hangul	"shkT"	"$(C3y(B"	kjhynmlop))

  (its-defrule-hangul-2	"shl"	"$(C3z(B"
    (its-define-hangul	"shls"	"$(C3{(B"	kijuhynbmlopP)
    (its-define-hangul	"shlf"	"$(C3|(B"	kijuhynbmlopP)
    (its-define-hangul	"shla"	"$(C3}(B"	kijuhynbmlopP)
    (its-define-hangul	"shlq"	"$(C3~(B"	kijuhynbmlopP)
    (its-define-hangul	"shlt"	"$(C4!(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"sy"	"$(C4"(B"
    (its-define-hangul	"syr"	"$(C4#(B"	kijuhynbmlopOP)
    (its-define-hangul	"sys"	"$(C4$(B"	kijuhynbmlopP)
    (its-define-hangul	"syf"	"$(C4%(B"	kijuhynbmlopP)
    (its-define-hangul	"syq"	"$(C4&(B"	kijuhynbmlopP)
    (its-define-hangul	"syt"	"$(C4'(B"	kijuhynbmlopOP)
    (its-define-hangul	"syd"	"$(C4((B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"sn"	"$(C4)(B"
    (its-define-hangul	"snr"	"$(C4*(B"	kijuhynbmlopOP)
    (its-define-hangul	"sns"	"$(C4+(B"	kijuhynbmlopP)
    (its-define-hangul	"sne"	"$(C4,(B"	kijuhynbmlopP)
    (its-define-hangul	"snf"	"$(C4-(B"	kijuhynbmlopP)
    (its-define-hangul	"sna"	"$(C4.(B"	kijuhynbmlopP)
    (its-define-hangul	"snq"	"$(C4/(B"	kijuhynbmlopP)
    (its-define-hangul	"snt"	"$(C40(B"	kijuhynbmlopOP)
    (its-define-hangul	"snd"	"$(C41(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"snj"	"$(C42(B"
    (its-define-hangul	"snjT"	"$(C43(B"	kjhynmlop))

  (its-defrule-hangul	"snp"	"$(C44(B")

  (its-defrule-hangul-2	"snl"	"$(C45(B"
    (its-define-hangul	"snls"	"$(C46(B"	kijuhynbmlopP)
    (its-define-hangul	"snlf"	"$(C47(B"	kijuhynbmlopP)
    (its-define-hangul	"snla"	"$(C48(B"	kijuhynbmlopP)
    (its-define-hangul	"snlq"	"$(C49(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sb"	"$(C4:(B"
    (its-define-hangul	"sbr"	"$(C4;(B"	kijuhynbmlopOP)
    (its-define-hangul	"sbf"	"$(C4<(B"	kijuhynbmlopP)
    (its-define-hangul	"sba"	"$(C4=(B"	kijuhynbmlopP)
    (its-define-hangul	"sbq"	"$(C4>(B"	kijuhynbmlopP)
    (its-define-hangul	"sbd"	"$(C4?(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"sm"	"$(C4@(B"
    (its-define-hangul	"smr"	"$(C4A(B"	kijuhynbmlopOP)
    (its-define-hangul	"sms"	"$(C4B(B"	kijuhynbmlopP)
    (its-define-hangul	"smf"	"$(C4C(B"	kijuhynbmlopP
     (its-define-hangul	"smfr"	"$(C4D(B"	kijuhynbmlopOP)
     (its-define-hangul	"smfa"	"$(C4E(B"	kijuhynbmlopP))
    (its-define-hangul	"sma"	"$(C4F(B"	kijuhynbmlopP)
    (its-define-hangul	"smq"	"$(C4G(B"	kijuhynbmlopP)
    (its-define-hangul	"smt"	"$(C4H(B"	kijuhynbmlopOP)
    (its-define-hangul	"smd"	"$(C4I(B"	kijuhynbmlopOP)
    (its-define-hangul	"smw"	"$(C4J(B"	kijuhynbmlopOP)
    (its-define-hangul	"smv"	"$(C4K(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sml"	"$(C4L(B"
    (its-define-hangul	"smls"	"$(C4M(B"	kijuhynbmlopP)
    (its-define-hangul	"smlf"	"$(C4N(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"sl"	"$(C4O(B"
    (its-define-hangul	"slr"	"$(C4P(B"	kijuhynbmlopOP)
    (its-define-hangul	"sls"	"$(C4Q(B"	kijuhynbmlopP)
    (its-define-hangul	"slf"	"$(C4R(B"	kijuhynbmlopP
     (its-define-hangul	"slfa"	"$(C4S(B"	kijuhynbmlopP))
    (its-define-hangul	"sla"	"$(C4T(B"	kijuhynbmlopP)
    (its-define-hangul	"slq"	"$(C4U(B"	kijuhynbmlopP)
    (its-define-hangul	"slt"	"$(C4V(B"	kijuhynbmlopOP)
    (its-define-hangul	"sld"	"$(C4W(B"	kijuhynbmlopOP)
    (its-define-hangul	"slv"	"$(C4X(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ek"	"$(C4Y(B"
    (its-define-hangul	"ekr"	"$(C4Z(B"	kijuhynbmlopOP)
    (its-define-hangul	"ekR"	"$(C4[(B"	kijuhynbmlopP)
    (its-define-hangul	"eks"	"$(C4\(B"	kijuhynbmlopP)
    (its-define-hangul	"eke"	"$(C4](B"	kijuhynbmlopP)
    (its-define-hangul	"ekf"	"$(C4^(B"	kijuhynbmlopP
     (its-define-hangul	"ekfr"	"$(C4_(B"	kijuhynbmlopOP)
     (its-define-hangul	"ekfa"	"$(C4`(B"	kijuhynbmlopP)
     (its-define-hangul	"ekfq"	"$(C4a(B"	kijuhynbmlopP)
     (its-define-hangul	"ekfg"	"$(C4b(B"	kijuhynbmlopP))
    (its-define-hangul	"eka"	"$(C4c(B"	kijuhynbmlopP)
    (its-define-hangul	"ekq"	"$(C4d(B"	kijuhynbmlopP)
    (its-define-hangul	"ekt"	"$(C4e(B"	kijuhynbmlopOP)
    (its-define-hangul	"ekT"	"$(C4f(B"	kjhynmlop)
    (its-define-hangul	"ekd"	"$(C4g(B"	kijuhynbmlopOP)
    (its-define-hangul	"ekw"	"$(C4h(B"	kijuhynbmlopOP)
    (its-define-hangul	"ekc"	"$(C4i(B"	kijuhynbmlopP)
    (its-define-hangul	"ekg"	"$(C4j(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"eo"	"$(C4k(B"
    (its-define-hangul	"eor"	"$(C4l(B"	kijuhynbmlopOP)
    (its-define-hangul	"eos"	"$(C4m(B"	kijuhynbmlopP)
    (its-define-hangul	"eof"	"$(C4n(B"	kijuhynbmlopP)
    (its-define-hangul	"eoa"	"$(C4o(B"	kijuhynbmlopP)
    (its-define-hangul	"eoq"	"$(C4p(B"	kijuhynbmlopP)
    (its-define-hangul	"eot"	"$(C4q(B"	kijuhynbmlopOP)
    (its-define-hangul	"eoT"	"$(C4r(B"	kjhynmlop)
    (its-define-hangul	"eod"	"$(C4s(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"ei"	"$(C4t(B")

  (its-defrule-hangul-2	"ej"	"$(C4u(B"
    (its-define-hangul	"ejr"	"$(C4v(B"	kijuhynbmlopOP)
    (its-define-hangul	"ejR"	"$(C4w(B"	kijuhynbmlopP)
    (its-define-hangul	"ejs"	"$(C4x(B"	kijuhynbmlopP)
    (its-define-hangul	"eje"	"$(C4y(B"	kijuhynbmlopP)
    (its-define-hangul	"ejf"	"$(C4z(B"	kijuhynbmlopP
     (its-define-hangul	"ejfa"	"$(C4{(B"	kijuhynbmlopP)
     (its-define-hangul	"ejfq"	"$(C4|(B"	kijuhynbmlopP))
    (its-define-hangul	"eja"	"$(C4}(B"	kijuhynbmlopP)
    (its-define-hangul	"ejq"	"$(C4~(B"	kijuhynbmlopP)
    (its-define-hangul	"ejt"	"$(C5!(B"	kijuhynbmlopOP)
    (its-define-hangul	"ejd"	"$(C5"(B"	kijuhynbmlopOP)
    (its-define-hangul	"ejc"	"$(C5#(B"	kijuhynbmlopP)
    (its-define-hangul	"ejv"	"$(C5$(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ep"	"$(C5%(B"
    (its-define-hangul	"epr"	"$(C5&(B"	kijuhynbmlopOP)
    (its-define-hangul	"eps"	"$(C5'(B"	kijuhynbmlopP)
    (its-define-hangul	"epf"	"$(C5((B"	kijuhynbmlopP)
    (its-define-hangul	"epa"	"$(C5)(B"	kijuhynbmlopP)
    (its-define-hangul	"epq"	"$(C5*(B"	kijuhynbmlopP)
    (its-define-hangul	"ept"	"$(C5+(B"	kijuhynbmlopOP)
    (its-define-hangul	"epT"	"$(C5,(B"	kjhynmlop)
    (its-define-hangul	"epd"	"$(C5-(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"eu"	"$(C5.(B"
    (its-define-hangul	"eus"	"$(C5/(B"	kijuhynbmlopP)
    (its-define-hangul	"euf"	"$(C50(B"	kijuhynbmlopP)
    (its-define-hangul	"euT"	"$(C51(B"	kjhynmlop)
    (its-define-hangul	"eud"	"$(C52(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"eP"	"$(C53(B"
    (its-define-hangul	"ePs"	"$(C54(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"eh"	"$(C55(B"
    (its-define-hangul	"ehr"	"$(C56(B"	kijuhynbmlopOP)
    (its-define-hangul	"ehs"	"$(C57(B"	kijuhynbmlopP)
    (its-define-hangul	"ehe"	"$(C58(B"	kijuhynbmlopP)
    (its-define-hangul	"ehf"	"$(C59(B"	kijuhynbmlopP
     (its-define-hangul	"ehfa"	"$(C5:(B"	kijuhynbmlopP)
     (its-define-hangul	"ehft"	"$(C5;(B"	kijuhynbmlopOP))
    (its-define-hangul	"eha"	"$(C5<(B"	kijuhynbmlopP)
    (its-define-hangul	"ehq"	"$(C5=(B"	kijuhynbmlopP)
    (its-define-hangul	"eht"	"$(C5>(B"	kijuhynbmlopOP)
    (its-define-hangul	"ehd"	"$(C5?(B"	kijuhynbmlopOP)
    (its-define-hangul	"ehc"	"$(C5@(B"	kijuhynbmlopP)
    (its-define-hangul	"ehx"	"$(C5A(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ehk"	"$(C5B(B"
    (its-define-hangul	"ehks"	"$(C5C(B"	kijuhynbmlopP)
    (its-define-hangul	"ehkf"	"$(C5D(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"eho"	"$(C5E(B"
    (its-define-hangul	"ehoT"	"$(C5F(B"	kjhynmlop))

  (its-defrule-hangul-2	"ehl"	"$(C5G(B"
    (its-define-hangul	"ehls"	"$(C5H(B"	kijuhynbmlopP)
    (its-define-hangul	"ehlf"	"$(C5I(B"	kijuhynbmlopP)
    (its-define-hangul	"ehla"	"$(C5J(B"	kijuhynbmlopP)
    (its-define-hangul	"ehlq"	"$(C5K(B"	kijuhynbmlopP)
    (its-define-hangul	"ehlt"	"$(C5L(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"ey"	"$(C5M(B")

  (its-defrule-hangul-2	"en"	"$(C5N(B"
    (its-define-hangul	"enr"	"$(C5O(B"	kijuhynbmlopOP)
    (its-define-hangul	"ens"	"$(C5P(B"	kijuhynbmlopP)
    (its-define-hangul	"enf"	"$(C5Q(B"	kijuhynbmlopP)
    (its-define-hangul	"ena"	"$(C5R(B"	kijuhynbmlopP)
    (its-define-hangul	"enq"	"$(C5S(B"	kijuhynbmlopP)
    (its-define-hangul	"ent"	"$(C5T(B"	kijuhynbmlopOP)
    (its-define-hangul	"end"	"$(C5U(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"enj"	"$(C5V(B"
    (its-define-hangul	"enjT"	"$(C5W(B"	kjhynmlop))

  (its-defrule-hangul-2	"enp"	"$(C5X(B"
    (its-define-hangul	"enpd"	"$(C5Y(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"enl"	"$(C5Z(B"
    (its-define-hangul	"enls"	"$(C5[(B"	kijuhynbmlopP)
    (its-define-hangul	"enlf"	"$(C5\(B"	kijuhynbmlopP)
    (its-define-hangul	"enlq"	"$(C5](B"	kijuhynbmlopP)
    (its-define-hangul	"enlt"	"$(C5^(B"	kijuhynbmlopOP)
    (its-define-hangul	"enld"	"$(C5_(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"eb"	"$(C5`(B"
    (its-define-hangul	"ebs"	"$(C5a(B"	kijuhynbmlopP)
    (its-define-hangul	"ebf"	"$(C5b(B"	kijuhynbmlopP)
    (its-define-hangul	"eba"	"$(C5c(B"	kijuhynbmlopP)
    (its-define-hangul	"ebd"	"$(C5d(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"em"	"$(C5e(B"
    (its-define-hangul	"emr"	"$(C5f(B"	kijuhynbmlopOP)
    (its-define-hangul	"ems"	"$(C5g(B"	kijuhynbmlopP)
    (its-define-hangul	"eme"	"$(C5h(B"	kijuhynbmlopP)
    (its-define-hangul	"emf"	"$(C5i(B"	kijuhynbmlopP
     (its-define-hangul	"emfa"	"$(C5j(B"	kijuhynbmlopP))
    (its-define-hangul	"ema"	"$(C5k(B"	kijuhynbmlopP)
    (its-define-hangul	"emq"	"$(C5l(B"	kijuhynbmlopP)
    (its-define-hangul	"emt"	"$(C5m(B"	kijuhynbmlopOP)
    (its-define-hangul	"emd"	"$(C5n(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"eml"	"$(C5o(B")

  (its-defrule-hangul-2	"el"	"$(C5p(B"
    (its-define-hangul	"elr"	"$(C5q(B"	kijuhynbmlopOP)
    (its-define-hangul	"els"	"$(C5r(B"	kijuhynbmlopP)
    (its-define-hangul	"ele"	"$(C5s(B"	kijuhynbmlopP)
    (its-define-hangul	"elf"	"$(C5t(B"	kijuhynbmlopP)
    (its-define-hangul	"ela"	"$(C5u(B"	kijuhynbmlopP)
    (its-define-hangul	"elq"	"$(C5v(B"	kijuhynbmlopP)
    (its-define-hangul	"elt"	"$(C5w(B"	kijuhynbmlopOP)
    (its-define-hangul	"elT"	"$(C5x(B"	kjhynmlop)
    (its-define-hangul	"eld"	"$(C5y(B"	kijuhynbmlopOP)
    (its-define-hangul	"elw"	"$(C5z(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Ek"	"$(C5{(B"
    (its-define-hangul	"Ekr"	"$(C5|(B"	kijuhynbmlopOP)
    (its-define-hangul	"Eks"	"$(C5}(B"	kijuhynbmlopP)
    (its-define-hangul	"Ekf"	"$(C5~(B"	kijuhynbmlopP)
    (its-define-hangul	"Eka"	"$(C6!(B"	kijuhynbmlopP)
    (its-define-hangul	"Ekq"	"$(C6"(B"	kijuhynbmlopP)
    (its-define-hangul	"Ekt"	"$(C6#(B"	kijuhynbmlopOP)
    (its-define-hangul	"EkT"	"$(C6$(B"	kjhynmlop)
    (its-define-hangul	"Ekd"	"$(C6%(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ekg"	"$(C6&(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Eo"	"$(C6'(B"
    (its-define-hangul	"Eor"	"$(C6((B"	kijuhynbmlopOP)
    (its-define-hangul	"Eos"	"$(C6)(B"	kijuhynbmlopP)
    (its-define-hangul	"Eof"	"$(C6*(B"	kijuhynbmlopP)
    (its-define-hangul	"Eoa"	"$(C6+(B"	kijuhynbmlopP)
    (its-define-hangul	"Eoq"	"$(C6,(B"	kijuhynbmlopP)
    (its-define-hangul	"Eot"	"$(C6-(B"	kijuhynbmlopOP)
    (its-define-hangul	"EoT"	"$(C6.(B"	kjhynmlop)
    (its-define-hangul	"Eod"	"$(C6/(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Ej"	"$(C60(B"
    (its-define-hangul	"Ejr"	"$(C61(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ejs"	"$(C62(B"	kijuhynbmlopP)
    (its-define-hangul	"Ejf"	"$(C63(B"	kijuhynbmlopP
     (its-define-hangul	"Ejfa"	"$(C64(B"	kijuhynbmlopP)
     (its-define-hangul	"Ejfq"	"$(C65(B"	kijuhynbmlopP))
    (its-define-hangul	"Eja"	"$(C66(B"	kijuhynbmlopP)
    (its-define-hangul	"Ejq"	"$(C67(B"	kijuhynbmlopP)
    (its-define-hangul	"Ejt"	"$(C68(B"	kijuhynbmlopOP)
    (its-define-hangul	"EjT"	"$(C69(B"	kjhynmlop)
    (its-define-hangul	"Ejd"	"$(C6:(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ejg"	"$(C6;(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Ep"	"$(C6<(B"
    (its-define-hangul	"Epr"	"$(C6=(B"	kijuhynbmlopOP)
    (its-define-hangul	"Eps"	"$(C6>(B"	kijuhynbmlopP)
    (its-define-hangul	"Epf"	"$(C6?(B"	kijuhynbmlopP)
    (its-define-hangul	"Epa"	"$(C6@(B"	kijuhynbmlopP)
    (its-define-hangul	"Epq"	"$(C6A(B"	kijuhynbmlopP)
    (its-define-hangul	"Ept"	"$(C6B(B"	kijuhynbmlopOP)
    (its-define-hangul	"EpT"	"$(C6C(B"	kjhynmlop)
    (its-define-hangul	"Epd"	"$(C6D(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Eu"	"$(C6E(B"
    (its-define-hangul	"EuT"	"$(C6F(B"	kjhynmlop))

  (its-defrule-hangul-2	"Eh"	"$(C6G(B"
    (its-define-hangul	"Ehr"	"$(C6H(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ehs"	"$(C6I(B"	kijuhynbmlopP)
    (its-define-hangul	"Ehf"	"$(C6J(B"	kijuhynbmlopP)
    (its-define-hangul	"Ehd"	"$(C6K(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Ehk"	"$(C6L(B"
    (its-define-hangul	"Ehkf"	"$(C6M(B"	kijuhynbmlopP))

  (its-defrule-hangul	"Eho"	"$(C6N(B")

  (its-defrule-hangul-2	"Ehl"	"$(C6O(B"
    (its-define-hangul	"Ehls"	"$(C6P(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"En"	"$(C6Q(B"
    (its-define-hangul	"Enr"	"$(C6R(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ens"	"$(C6S(B"	kijuhynbmlopP)
    (its-define-hangul	"Enf"	"$(C6T(B"	kijuhynbmlopP
     (its-define-hangul	"Enfg"	"$(C6U(B"	kijuhynbmlopP))
    (its-define-hangul	"Ena"	"$(C6V(B"	kijuhynbmlopP)
    (its-define-hangul	"End"	"$(C6W(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"Enp"	"$(C6X(B")

  (its-defrule-hangul-2	"Enl"	"$(C6Y(B"
    (its-define-hangul	"Enls"	"$(C6Z(B"	kijuhynbmlopP)
    (its-define-hangul	"Enlf"	"$(C6[(B"	kijuhynbmlopP)
    (its-define-hangul	"Enla"	"$(C6\(B"	kijuhynbmlopP)
    (its-define-hangul	"Enlq"	"$(C6](B"	kijuhynbmlopP)
    (its-define-hangul	"Enld"	"$(C6^(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Em"	"$(C6_(B"
    (its-define-hangul	"Emr"	"$(C6`(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ems"	"$(C6a(B"	kijuhynbmlopP)
    (its-define-hangul	"Eme"	"$(C6b(B"	kijuhynbmlopP)
    (its-define-hangul	"Emf"	"$(C6c(B"	kijuhynbmlopP)
    (its-define-hangul	"Ema"	"$(C6d(B"	kijuhynbmlopP)
    (its-define-hangul	"Emq"	"$(C6e(B"	kijuhynbmlopP)
    (its-define-hangul	"Emt"	"$(C6f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Eml"	"$(C6g(B"
    (its-define-hangul	"Emls"	"$(C6h(B"	kijuhynbmlopP)
    (its-define-hangul	"Emlf"	"$(C6i(B"	kijuhynbmlopP)
    (its-define-hangul	"Emla"	"$(C6j(B"	kijuhynbmlopP)
    (its-define-hangul	"Emlq"	"$(C6k(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"El"	"$(C6l(B"
    (its-define-hangul	"Els"	"$(C6m(B"	kijuhynbmlopP)
    (its-define-hangul	"Elf"	"$(C6n(B"	kijuhynbmlopP)
    (its-define-hangul	"Ela"	"$(C6o(B"	kijuhynbmlopP)
    (its-define-hangul	"Elq"	"$(C6p(B"	kijuhynbmlopP)
    (its-define-hangul	"Elt"	"$(C6q(B"	kijuhynbmlopOP)
    (its-define-hangul	"Eld"	"$(C6r(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fk"	"$(C6s(B"
    (its-define-hangul	"fkr"	"$(C6t(B"	kijuhynbmlopOP)
    (its-define-hangul	"fks"	"$(C6u(B"	kijuhynbmlopP)
    (its-define-hangul	"fkf"	"$(C6v(B"	kijuhynbmlopP)
    (its-define-hangul	"fka"	"$(C6w(B"	kijuhynbmlopP)
    (its-define-hangul	"fkq"	"$(C6x(B"	kijuhynbmlopP)
    (its-define-hangul	"fkt"	"$(C6y(B"	kijuhynbmlopOP)
    (its-define-hangul	"fkT"	"$(C6z(B"	kjhynmlop)
    (its-define-hangul	"fkd"	"$(C6{(B"	kijuhynbmlopOP)
    (its-define-hangul	"fkw"	"$(C6|(B"	kijuhynbmlopOP)
    (its-define-hangul	"fkv"	"$(C6}(B"	kijuhynbmlopP)
    (its-define-hangul	"fkg"	"$(C6~(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"fo"	"$(C7!(B"
    (its-define-hangul	"for"	"$(C7"(B"	kijuhynbmlopOP)
    (its-define-hangul	"fos"	"$(C7#(B"	kijuhynbmlopP)
    (its-define-hangul	"fof"	"$(C7$(B"	kijuhynbmlopP)
    (its-define-hangul	"foa"	"$(C7%(B"	kijuhynbmlopP)
    (its-define-hangul	"foq"	"$(C7&(B"	kijuhynbmlopP)
    (its-define-hangul	"fot"	"$(C7'(B"	kijuhynbmlopOP)
    (its-define-hangul	"foT"	"$(C7((B"	kjhynmlop)
    (its-define-hangul	"fod"	"$(C7)(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fi"	"$(C7*(B"
    (its-define-hangul	"fir"	"$(C7+(B"	kijuhynbmlopOP)
    (its-define-hangul	"fis"	"$(C7,(B"	kijuhynbmlopP)
    (its-define-hangul	"fit"	"$(C7-(B"	kijuhynbmlopOP)
    (its-define-hangul	"fid"	"$(C7.(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fj"	"$(C7/(B"
    (its-define-hangul	"fjr"	"$(C70(B"	kijuhynbmlopOP)
    (its-define-hangul	"fjs"	"$(C71(B"	kijuhynbmlopP)
    (its-define-hangul	"fjf"	"$(C72(B"	kijuhynbmlopP)
    (its-define-hangul	"fja"	"$(C73(B"	kijuhynbmlopP)
    (its-define-hangul	"fjq"	"$(C74(B"	kijuhynbmlopP)
    (its-define-hangul	"fjt"	"$(C75(B"	kijuhynbmlopOP)
    (its-define-hangul	"fjT"	"$(C76(B"	kjhynmlop)
    (its-define-hangul	"fjd"	"$(C77(B"	kijuhynbmlopOP)
    (its-define-hangul	"fjg"	"$(C78(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"fp"	"$(C79(B"
    (its-define-hangul	"fpr"	"$(C7:(B"	kijuhynbmlopOP)
    (its-define-hangul	"fps"	"$(C7;(B"	kijuhynbmlopP)
    (its-define-hangul	"fpf"	"$(C7<(B"	kijuhynbmlopP)
    (its-define-hangul	"fpa"	"$(C7=(B"	kijuhynbmlopP)
    (its-define-hangul	"fpq"	"$(C7>(B"	kijuhynbmlopP)
    (its-define-hangul	"fpt"	"$(C7?(B"	kijuhynbmlopOP)
    (its-define-hangul	"fpd"	"$(C7@(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fu"	"$(C7A(B"
    (its-define-hangul	"fur"	"$(C7B(B"	kijuhynbmlopOP)
    (its-define-hangul	"fus"	"$(C7C(B"	kijuhynbmlopP)
    (its-define-hangul	"fuf"	"$(C7D(B"	kijuhynbmlopP)
    (its-define-hangul	"fua"	"$(C7E(B"	kijuhynbmlopP)
    (its-define-hangul	"fuq"	"$(C7F(B"	kijuhynbmlopP)
    (its-define-hangul	"fut"	"$(C7G(B"	kijuhynbmlopOP)
    (its-define-hangul	"fuT"	"$(C7H(B"	kjhynmlop)
    (its-define-hangul	"fud"	"$(C7I(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fP"	"$(C7J(B"
    (its-define-hangul	"fPs"	"$(C7K(B"	kijuhynbmlopP)
    (its-define-hangul	"fPq"	"$(C7L(B"	kijuhynbmlopP)
    (its-define-hangul	"fPt"	"$(C7M(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fh"	"$(C7N(B"
    (its-define-hangul	"fhr"	"$(C7O(B"	kijuhynbmlopOP)
    (its-define-hangul	"fhs"	"$(C7P(B"	kijuhynbmlopP)
    (its-define-hangul	"fhf"	"$(C7Q(B"	kijuhynbmlopP)
    (its-define-hangul	"fha"	"$(C7R(B"	kijuhynbmlopP)
    (its-define-hangul	"fhq"	"$(C7S(B"	kijuhynbmlopP)
    (its-define-hangul	"fht"	"$(C7T(B"	kijuhynbmlopOP)
    (its-define-hangul	"fhd"	"$(C7U(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fhk"	"$(C7V(B"
    (its-define-hangul	"fhks"	"$(C7W(B"	kijuhynbmlopP)
    (its-define-hangul	"fhkd"	"$(C7X(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"fho"	"$(C7N(Bo")
  (its-defrule-hangul	"fhoT"	"$(C7Y(B")

  (its-defrule-hangul-2	"fhl"	"$(C7Z(B"
    (its-define-hangul	"fhls"	"$(C7[(B"	kijuhynbmlopP)
    (its-define-hangul	"fhlf"	"$(C7\(B"	kijuhynbmlopP)
    (its-define-hangul	"fhla"	"$(C7](B"	kijuhynbmlopP)
    (its-define-hangul	"fhlq"	"$(C7^(B"	kijuhynbmlopP)
    (its-define-hangul	"fhlt"	"$(C7_(B"	kijuhynbmlopOP)
    (its-define-hangul	"fhld"	"$(C7`(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fy"	"$(C7a(B"
    (its-define-hangul	"fys"	"$(C7b(B"	kijuhynbmlopP)
    (its-define-hangul	"fyf"	"$(C7c(B"	kijuhynbmlopP)
    (its-define-hangul	"fyq"	"$(C7d(B"	kijuhynbmlopP)
    (its-define-hangul	"fyt"	"$(C7e(B"	kijuhynbmlopOP)
    (its-define-hangul	"fyd"	"$(C7f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fn"	"$(C7g(B"
    (its-define-hangul	"fnr"	"$(C7h(B"	kijuhynbmlopOP)
    (its-define-hangul	"fns"	"$(C7i(B"	kijuhynbmlopP)
    (its-define-hangul	"fnf"	"$(C7j(B"	kijuhynbmlopP)
    (its-define-hangul	"fna"	"$(C7k(B"	kijuhynbmlopP)
    (its-define-hangul	"fnq"	"$(C7l(B"	kijuhynbmlopP)
    (its-define-hangul	"fnt"	"$(C7m(B"	kijuhynbmlopOP)
    (its-define-hangul	"fnd"	"$(C7n(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fnj"	"$(C7o(B"
    (its-define-hangul	"fnjT"	"$(C7p(B"	kjhynmlop))

  (its-defrule-hangul	"fnp"	"$(C7q(B")

  (its-defrule-hangul-2	"fnl"	"$(C7r(B"
    (its-define-hangul	"fnlr"	"$(C7s(B"	kijuhynbmlopOP)
    (its-define-hangul	"fnls"	"$(C7t(B"	kijuhynbmlopP)
    (its-define-hangul	"fnlf"	"$(C7u(B"	kijuhynbmlopP)
    (its-define-hangul	"fnla"	"$(C7v(B"	kijuhynbmlopP)
    (its-define-hangul	"fnlt"	"$(C7w(B"	kijuhynbmlopOP)
    (its-define-hangul	"fnld"	"$(C7x(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fb"	"$(C7y(B"
    (its-define-hangul	"fbr"	"$(C7z(B"	kijuhynbmlopOP)
    (its-define-hangul	"fbs"	"$(C7{(B"	kijuhynbmlopP)
    (its-define-hangul	"fbf"	"$(C7|(B"	kijuhynbmlopP)
    (its-define-hangul	"fba"	"$(C7}(B"	kijuhynbmlopP)
    (its-define-hangul	"fbq"	"$(C7~(B"	kijuhynbmlopP)
    (its-define-hangul	"fbt"	"$(C8!(B"	kijuhynbmlopOP)
    (its-define-hangul	"fbd"	"$(C8"(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"fm"	"$(C8#(B"
    (its-define-hangul	"fmr"	"$(C8$(B"	kijuhynbmlopOP)
    (its-define-hangul	"fms"	"$(C8%(B"	kijuhynbmlopP)
    (its-define-hangul	"fmf"	"$(C8&(B"	kijuhynbmlopP)
    (its-define-hangul	"fma"	"$(C8'(B"	kijuhynbmlopP)
    (its-define-hangul	"fmq"	"$(C8((B"	kijuhynbmlopP)
    (its-define-hangul	"fmt"	"$(C8)(B"	kijuhynbmlopOP)
    (its-define-hangul	"fmd"	"$(C8*(B"	kijuhynbmlopOP)
    (its-define-hangul	"fmw"	"$(C8+(B"	kijuhynbmlopOP)
    (its-define-hangul	"fmx"	"$(C8,(B"	kijuhynbmlopP)
    (its-define-hangul	"fmv"	"$(C8-(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"fl"	"$(C8.(B"
    (its-define-hangul	"flr"	"$(C8/(B"	kijuhynbmlopOP)
    (its-define-hangul	"fls"	"$(C80(B"	kijuhynbmlopP)
    (its-define-hangul	"flf"	"$(C81(B"	kijuhynbmlopP)
    (its-define-hangul	"fla"	"$(C82(B"	kijuhynbmlopP)
    (its-define-hangul	"flq"	"$(C83(B"	kijuhynbmlopP)
    (its-define-hangul	"flt"	"$(C84(B"	kijuhynbmlopOP)
    (its-define-hangul	"fld"	"$(C85(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ak"	"$(C86(B"
    (its-define-hangul	"akr"	"$(C87(B"	kijuhynbmlopOP)
    (its-define-hangul	"aks"	"$(C88(B"	kijuhynbmlopP
     (its-define-hangul	"aksg"	"$(C89(B"	kijuhynbmlopP))
    (its-define-hangul	"ake"	"$(C8:(B"	kijuhynbmlopP)
    (its-define-hangul	"akf"	"$(C8;(B"	kijuhynbmlopP
     (its-define-hangul	"akfr"	"$(C8<(B"	kijuhynbmlopOP)
     (its-define-hangul	"akfa"	"$(C8=(B"	kijuhynbmlopP))
    (its-define-hangul	"aka"	"$(C8>(B"	kijuhynbmlopP)
    (its-define-hangul	"akq"	"$(C8?(B"	kijuhynbmlopP)
    (its-define-hangul	"akt"	"$(C8@(B"	kijuhynbmlopOP)
    (its-define-hangul	"akd"	"$(C8A(B"	kijuhynbmlopOP)
    (its-define-hangul	"akw"	"$(C8B(B"	kijuhynbmlopOP)
    (its-define-hangul	"akx"	"$(C8C(B"	kijuhynbmlopP)
    (its-define-hangul	"akg"	"$(C8D(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ao"	"$(C8E(B"
    (its-define-hangul	"aor"	"$(C8F(B"	kijuhynbmlopOP)
    (its-define-hangul	"aos"	"$(C8G(B"	kijuhynbmlopP)
    (its-define-hangul	"aof"	"$(C8H(B"	kijuhynbmlopP)
    (its-define-hangul	"aoa"	"$(C8I(B"	kijuhynbmlopP)
    (its-define-hangul	"aoq"	"$(C8J(B"	kijuhynbmlopP)
    (its-define-hangul	"aot"	"$(C8K(B"	kijuhynbmlopOP)
    (its-define-hangul	"aoT"	"$(C8L(B"	kjhynmlop)
    (its-define-hangul	"aod"	"$(C8M(B"	kijuhynbmlopOP)
    (its-define-hangul	"aow"	"$(C8N(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ai"	"$(C8O(B"
    (its-define-hangul	"air"	"$(C8P(B"	kijuhynbmlopOP)
    (its-define-hangul	"aif"	"$(C8Q(B"	kijuhynbmlopP)
    (its-define-hangul	"aid"	"$(C8R(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"aj"	"$(C8S(B"
    (its-define-hangul	"ajr"	"$(C8T(B"	kijuhynbmlopOP)
    (its-define-hangul	"ajs"	"$(C8U(B"	kijuhynbmlopP)
    (its-define-hangul	"ajf"	"$(C8V(B"	kijuhynbmlopP
     (its-define-hangul	"ajfa"	"$(C8W(B"	kijuhynbmlopP))
    (its-define-hangul	"aja"	"$(C8X(B"	kijuhynbmlopP)
    (its-define-hangul	"ajq"	"$(C8Y(B"	kijuhynbmlopP)
    (its-define-hangul	"ajt"	"$(C8Z(B"	kijuhynbmlopOP)
    (its-define-hangul	"ajd"	"$(C8[(B"	kijuhynbmlopOP)
    (its-define-hangul	"ajw"	"$(C8\(B"	kijuhynbmlopOP)
    (its-define-hangul	"ajg"	"$(C8](B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ap"	"$(C8^(B"
    (its-define-hangul	"apr"	"$(C8_(B"	kijuhynbmlopOP)
    (its-define-hangul	"aps"	"$(C8`(B"	kijuhynbmlopP)
    (its-define-hangul	"apf"	"$(C8a(B"	kijuhynbmlopP)
    (its-define-hangul	"apa"	"$(C8b(B"	kijuhynbmlopP)
    (its-define-hangul	"apq"	"$(C8c(B"	kijuhynbmlopP)
    (its-define-hangul	"apt"	"$(C8d(B"	kijuhynbmlopOP)
    (its-define-hangul	"apT"	"$(C8e(B"	kjhynmlop)
    (its-define-hangul	"apd"	"$(C8f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"au"	"$(C8g(B"
    (its-define-hangul	"aur"	"$(C8h(B"	kijuhynbmlopOP)
    (its-define-hangul	"aus"	"$(C8i(B"	kijuhynbmlopP)
    (its-define-hangul	"auf"	"$(C8j(B"	kijuhynbmlopP)
    (its-define-hangul	"aut"	"$(C8k(B"	kijuhynbmlopOP)
    (its-define-hangul	"auT"	"$(C8l(B"	kjhynmlop)
    (its-define-hangul	"aud"	"$(C8m(B"	kijuhynbmlopOP)
    (its-define-hangul	"auc"	"$(C8n(B"	kijuhynbmlopP))

  (its-defrule-hangul	"aP"	"$(C8o(B")

  (its-defrule-hangul-2	"ah"	"$(C8p(B"
    (its-define-hangul	"ahr"	"$(C8q(B"	kijuhynbmlopOP
     (its-define-hangul	"ahrt"	"$(C8r(B"	kijuhynbmlopOP))
    (its-define-hangul	"ahs"	"$(C8s(B"	kijuhynbmlopP)
    (its-define-hangul	"ahf"	"$(C8t(B"	kijuhynbmlopP
     (its-define-hangul	"ahfa"	"$(C8u(B"	kijuhynbmlopP))
    (its-define-hangul	"aha"	"$(C8v(B"	kijuhynbmlopP)
    (its-define-hangul	"ahq"	"$(C8w(B"	kijuhynbmlopP)
    (its-define-hangul	"aht"	"$(C8x(B"	kijuhynbmlopOP)
    (its-define-hangul	"ahd"	"$(C8y(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ahk"	"$(C8z(B"
    (its-define-hangul	"ahks"	"$(C8{(B"	kijuhynbmlopP)
    (its-define-hangul	"ahkT"	"$(C8|(B"	kjhynmlop)
    (its-define-hangul	"ahkd"	"$(C8}(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ahl"	"$(C8~(B"
    (its-define-hangul	"ahls"	"$(C9!(B"	kijuhynbmlopP)
    (its-define-hangul	"ahlf"	"$(C9"(B"	kijuhynbmlopP)
    (its-define-hangul	"ahlq"	"$(C9#(B"	kijuhynbmlopP)
    (its-define-hangul	"ahlt"	"$(C9$(B"	kijuhynbmlopOP)
    (its-define-hangul	"ahld"	"$(C9%(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ay"	"$(C9&(B"
    (its-define-hangul	"ays"	"$(C9'(B"	kijuhynbmlopP)
    (its-define-hangul	"ayf"	"$(C9((B"	kijuhynbmlopP)
    (its-define-hangul	"ayq"	"$(C9)(B"	kijuhynbmlopP)
    (its-define-hangul	"ayt"	"$(C9*(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"an"	"$(C9+(B"
    (its-define-hangul	"anr"	"$(C9,(B"	kijuhynbmlopOP)
    (its-define-hangul	"anR"	"$(C9-(B"	kijuhynbmlopP)
    (its-define-hangul	"ans"	"$(C9.(B"	kijuhynbmlopP)
    (its-define-hangul	"ane"	"$(C9/(B"	kijuhynbmlopP)
    (its-define-hangul	"anf"	"$(C90(B"	kijuhynbmlopP
     (its-define-hangul	"anfr"	"$(C91(B"	kijuhynbmlopOP)
     (its-define-hangul	"anfa"	"$(C92(B"	kijuhynbmlopP))
    (its-define-hangul	"ana"	"$(C93(B"	kijuhynbmlopP)
    (its-define-hangul	"anq"	"$(C94(B"	kijuhynbmlopP)
    (its-define-hangul	"ant"	"$(C95(B"	kijuhynbmlopOP)
    (its-define-hangul	"and"	"$(C96(B"	kijuhynbmlopOP)
    (its-define-hangul	"anx"	"$(C97(B"	kijuhynbmlopP)
    (its-define-hangul	"ang"	"$(C98(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"anj"	"$(C99(B"
    (its-define-hangul	"anjs"	"$(C9:(B"	kijuhynbmlopP)
    (its-define-hangul	"anjf"	"$(C9;(B"	kijuhynbmlopP)
    (its-define-hangul	"anjq"	"$(C9<(B"	kijuhynbmlopP)
    (its-define-hangul	"anjt"	"$(C9=(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"anp"	"$(C9>(B")

  (its-defrule-hangul-2	"anl"	"$(C9?(B"
    (its-define-hangul	"anls"	"$(C9@(B"	kijuhynbmlopP)
    (its-define-hangul	"anlf"	"$(C9A(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ab"	"$(C9B(B"
    (its-define-hangul	"abs"	"$(C9C(B"	kijuhynbmlopP)
    (its-define-hangul	"abf"	"$(C9D(B"	kijuhynbmlopP)
    (its-define-hangul	"aba"	"$(C9E(B"	kijuhynbmlopP)
    (its-define-hangul	"abt"	"$(C9F(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"am"	"$(C9G(B"
    (its-define-hangul	"ams"	"$(C9H(B"	kijuhynbmlopP)
    (its-define-hangul	"amf"	"$(C9I(B"	kijuhynbmlopP)
    (its-define-hangul	"ama"	"$(C9J(B"	kijuhynbmlopP)
    (its-define-hangul	"amt"	"$(C9K(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"al"	"$(C9L(B"
    (its-define-hangul	"alr"	"$(C9M(B"	kijuhynbmlopOP)
    (its-define-hangul	"als"	"$(C9N(B"	kijuhynbmlopP)
    (its-define-hangul	"ale"	"$(C9O(B"	kijuhynbmlopP)
    (its-define-hangul	"alf"	"$(C9P(B"	kijuhynbmlopP
     (its-define-hangul	"alfa"	"$(C9Q(B"	kijuhynbmlopP))
    (its-define-hangul	"ala"	"$(C9R(B"	kijuhynbmlopP)
    (its-define-hangul	"alq"	"$(C9S(B"	kijuhynbmlopP)
    (its-define-hangul	"alt"	"$(C9T(B"	kijuhynbmlopOP)
    (its-define-hangul	"alT"	"$(C9U(B"	kjhynmlop)
    (its-define-hangul	"ald"	"$(C9V(B"	kijuhynbmlopOP)
    (its-define-hangul	"alc"	"$(C9W(B"	kijuhynbmlopP)
    (its-define-hangul	"alx"	"$(C9X(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qk"	"$(C9Y(B"
    (its-define-hangul	"qkr"	"$(C9Z(B"	kijuhynbmlopOP
     (its-define-hangul	"qkrt"	"$(C9\(B"	kijuhynbmlopOP))
    (its-define-hangul	"qkR"	"$(C9[(B"	kijuhynbmlopP)
    (its-define-hangul	"qks"	"$(C9](B"	kijuhynbmlopP)
    (its-define-hangul	"qke"	"$(C9^(B"	kijuhynbmlopP)
    (its-define-hangul	"qkf"	"$(C9_(B"	kijuhynbmlopP
     (its-define-hangul	"qkfr"	"$(C9`(B"	kijuhynbmlopOP)
     (its-define-hangul	"qkfa"	"$(C9a(B"	kijuhynbmlopP)
     (its-define-hangul	"qkfq"	"$(C9b(B"	kijuhynbmlopP))
    (its-define-hangul	"qka"	"$(C9c(B"	kijuhynbmlopP)
    (its-define-hangul	"qkq"	"$(C9d(B"	kijuhynbmlopP)
    (its-define-hangul	"qkt"	"$(C9e(B"	kijuhynbmlopOP)
    (its-define-hangul	"qkd"	"$(C9f(B"	kijuhynbmlopOP)
    (its-define-hangul	"qkx"	"$(C9g(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qo"	"$(C9h(B"
    (its-define-hangul	"qor"	"$(C9i(B"	kijuhynbmlopOP)
    (its-define-hangul	"qos"	"$(C9j(B"	kijuhynbmlopP)
    (its-define-hangul	"qof"	"$(C9k(B"	kijuhynbmlopP)
    (its-define-hangul	"qoa"	"$(C9l(B"	kijuhynbmlopP)
    (its-define-hangul	"qoq"	"$(C9m(B"	kijuhynbmlopP)
    (its-define-hangul	"qot"	"$(C9n(B"	kijuhynbmlopOP)
    (its-define-hangul	"qoT"	"$(C9o(B"	kjhynmlop)
    (its-define-hangul	"qod"	"$(C9p(B"	kijuhynbmlopOP)
    (its-define-hangul	"qox"	"$(C9q(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qi"	"$(C9r(B"
    (its-define-hangul	"qir"	"$(C9s(B"	kijuhynbmlopOP)
    (its-define-hangul	"qis"	"$(C9t(B"	kijuhynbmlopP)
    (its-define-hangul	"qiq"	"$(C9u(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qj"	"$(C9v(B"
    (its-define-hangul	"qjr"	"$(C9w(B"	kijuhynbmlopOP)
    (its-define-hangul	"qjs"	"$(C9x(B"	kijuhynbmlopP)
    (its-define-hangul	"qje"	"$(C9y(B"	kijuhynbmlopP)
    (its-define-hangul	"qjf"	"$(C9z(B"	kijuhynbmlopP
     (its-define-hangul	"qjfa"	"$(C9{(B"	kijuhynbmlopP))
    (its-define-hangul	"qja"	"$(C9|(B"	kijuhynbmlopP)
    (its-define-hangul	"qjq"	"$(C9}(B"	kijuhynbmlopP)
    (its-define-hangul	"qjt"	"$(C9~(B"	kijuhynbmlopOP)
    (its-define-hangul	"qjd"	"$(C:!(B"	kijuhynbmlopOP)
    (its-define-hangul	"qjw"	"$(C:"(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"qp"	"$(C:#(B"
    (its-define-hangul	"qpr"	"$(C:$(B"	kijuhynbmlopOP)
    (its-define-hangul	"qps"	"$(C:%(B"	kijuhynbmlopP)
    (its-define-hangul	"qpe"	"$(C:&(B"	kijuhynbmlopP)
    (its-define-hangul	"qpf"	"$(C:'(B"	kijuhynbmlopP)
    (its-define-hangul	"qpa"	"$(C:((B"	kijuhynbmlopP)
    (its-define-hangul	"qpq"	"$(C:)(B"	kijuhynbmlopP)
    (its-define-hangul	"qpt"	"$(C:*(B"	kijuhynbmlopOP)
    (its-define-hangul	"qpT"	"$(C:+(B"	kjhynmlop)
    (its-define-hangul	"qpd"	"$(C:,(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"qu"	"$(C:-(B"
    (its-define-hangul	"qur"	"$(C:.(B"	kijuhynbmlopOP)
    (its-define-hangul	"qus"	"$(C:/(B"	kijuhynbmlopP)
    (its-define-hangul	"quf"	"$(C:0(B"	kijuhynbmlopP)
    (its-define-hangul	"quq"	"$(C:1(B"	kijuhynbmlopP)
    (its-define-hangul	"qut"	"$(C:2(B"	kijuhynbmlopOP)
    (its-define-hangul	"quT"	"$(C:3(B"	kjhynmlop)
    (its-define-hangul	"qud"	"$(C:4(B"	kijuhynbmlopOP)
    (its-define-hangul	"qux"	"$(C:5(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qP"	"$(C:6(B"
    (its-define-hangul	"qPs"	"$(C:7(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qh"	"$(C:8(B"
    (its-define-hangul	"qhr"	"$(C:9(B"	kijuhynbmlopOP)
    (its-define-hangul	"qhR"	"$(C::(B"	kijuhynbmlopP)
    (its-define-hangul	"qhs"	"$(C:;(B"	kijuhynbmlopP)
    (its-define-hangul	"qhf"	"$(C:<(B"	kijuhynbmlopP)
    (its-define-hangul	"qha"	"$(C:=(B"	kijuhynbmlopP)
    (its-define-hangul	"qhq"	"$(C:>(B"	kijuhynbmlopP)
    (its-define-hangul	"qht"	"$(C:?(B"	kijuhynbmlopOP)
    (its-define-hangul	"qhd"	"$(C:@(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"qhk"	"$(C:A(B"
    (its-define-hangul	"qhks"	"$(C:B(B"	kijuhynbmlopP)
    (its-define-hangul	"qhkT"	"$(C:C(B"	kjhynmlop))

  (its-defrule-hangul-2	"qho"	"$(C:D(B"
    (its-define-hangul	"qhoT"	"$(C:E(B"	kjhynmlop))

  (its-defrule-hangul-2	"qhl"	"$(C:F(B"
    (its-define-hangul	"qhlr"	"$(C:G(B"	kijuhynbmlopOP)
    (its-define-hangul	"qhls"	"$(C:H(B"	kijuhynbmlopP)
    (its-define-hangul	"qhlf"	"$(C:I(B"	kijuhynbmlopP)
    (its-define-hangul	"qhla"	"$(C:J(B"	kijuhynbmlopP)
    (its-define-hangul	"qhlq"	"$(C:K(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qy"	"$(C:L(B"
    (its-define-hangul	"qys"	"$(C:M(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qn"	"$(C:N(B"
    (its-define-hangul	"qnr"	"$(C:O(B"	kijuhynbmlopOP)
    (its-define-hangul	"qns"	"$(C:P(B"	kijuhynbmlopP)
    (its-define-hangul	"qne"	"$(C:Q(B"	kijuhynbmlopP)
    (its-define-hangul	"qnf"	"$(C:R(B"	kijuhynbmlopP
     (its-define-hangul	"qnfr"	"$(C:S(B"	kijuhynbmlopOP)
     (its-define-hangul	"qnfa"	"$(C:T(B"	kijuhynbmlopP))
    (its-define-hangul	"qna"	"$(C:U(B"	kijuhynbmlopP)
    (its-define-hangul	"qnq"	"$(C:V(B"	kijuhynbmlopP)
    (its-define-hangul	"qnt"	"$(C:W(B"	kijuhynbmlopOP)
    (its-define-hangul	"qnd"	"$(C:X(B"	kijuhynbmlopOP)
    (its-define-hangul	"qnx"	"$(C:Y(B"	kijuhynbmlopP)
    (its-define-hangul	"qnv"	"$(C:Z(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"qnj"	"$(C:[(B"
    (its-define-hangul	"qnjf"	"$(C:\(B"	kijuhynbmlopP)
    (its-define-hangul	"qnjT"	"$(C:](B"	kjhynmlop))

  (its-defrule-hangul	"qnp"	"$(C:^(B")

  (its-defrule-hangul-2	"qnl"	"$(C:_(B"
    (its-define-hangul	"qnlr"	"$(C:`(B"	kijuhynbmlopOP)
    (its-define-hangul	"qnls"	"$(C:a(B"	kijuhynbmlopP)
    (its-define-hangul	"qnlf"	"$(C:b(B"	kijuhynbmlopP)
    (its-define-hangul	"qnld"	"$(C:c(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"qb"	"$(C:d(B"
    (its-define-hangul	"qbs"	"$(C:e(B"	kijuhynbmlopP)
    (its-define-hangul	"qbf"	"$(C:f(B"	kijuhynbmlopP)
    (its-define-hangul	"qba"	"$(C:g(B"	kijuhynbmlopP)
    (its-define-hangul	"qbt"	"$(C:h(B"	kijuhynbmlopOP)
    (its-define-hangul	"qbd"	"$(C:i(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"qm"	"$(C:j(B"
    (its-define-hangul	"qmr"	"$(C:k(B"	kijuhynbmlopOP)
    (its-define-hangul	"qms"	"$(C:l(B"	kijuhynbmlopP)
    (its-define-hangul	"qmf"	"$(C:m(B"	kijuhynbmlopP)
    (its-define-hangul	"qma"	"$(C:n(B"	kijuhynbmlopP)
    (its-define-hangul	"qmq"	"$(C:o(B"	kijuhynbmlopP)
    (its-define-hangul	"qmt"	"$(C:p(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ql"	"$(C:q(B"
    (its-define-hangul	"qlr"	"$(C:r(B"	kijuhynbmlopOP)
    (its-define-hangul	"qls"	"$(C:s(B"	kijuhynbmlopP)
    (its-define-hangul	"qlf"	"$(C:t(B"	kijuhynbmlopP
     (its-define-hangul	"qlfa"	"$(C:u(B"	kijuhynbmlopP))
    (its-define-hangul	"qla"	"$(C:v(B"	kijuhynbmlopP)
    (its-define-hangul	"qlq"	"$(C:w(B"	kijuhynbmlopP)
    (its-define-hangul	"qlt"	"$(C:x(B"	kijuhynbmlopOP)
    (its-define-hangul	"qld"	"$(C:y(B"	kijuhynbmlopOP)
    (its-define-hangul	"qlw"	"$(C:z(B"	kijuhynbmlopOP)
    (its-define-hangul	"qlc"	"$(C:{(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Qk"	"$(C:|(B"
    (its-define-hangul	"Qkr"	"$(C:}(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qks"	"$(C:~(B"	kijuhynbmlopP)
    (its-define-hangul	"Qkf"	"$(C;!(B"	kijuhynbmlopP
     (its-define-hangul	"Qkfa"	"$(C;"(B"	kijuhynbmlopP))
    (its-define-hangul	"Qka"	"$(C;#(B"	kijuhynbmlopP)
    (its-define-hangul	"Qkq"	"$(C;$(B"	kijuhynbmlopP)
    (its-define-hangul	"Qkt"	"$(C;%(B"	kijuhynbmlopOP)
    (its-define-hangul	"QkT"	"$(C;&(B"	kjhynmlop)
    (its-define-hangul	"Qkd"	"$(C;'(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qkg"	"$(C;((B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Qo"	"$(C;)(B"
    (its-define-hangul	"Qor"	"$(C;*(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qos"	"$(C;+(B"	kijuhynbmlopP)
    (its-define-hangul	"Qof"	"$(C;,(B"	kijuhynbmlopP)
    (its-define-hangul	"Qoa"	"$(C;-(B"	kijuhynbmlopP)
    (its-define-hangul	"Qoq"	"$(C;.(B"	kijuhynbmlopP)
    (its-define-hangul	"Qot"	"$(C;/(B"	kijuhynbmlopOP)
    (its-define-hangul	"QoT"	"$(C;0(B"	kjhynmlop)
    (its-define-hangul	"Qod"	"$(C;1(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qi"	"$(C;2(B"
    (its-define-hangul	"Qir"	"$(C;3(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qia"	"$(C;4(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Qj"	"$(C;5(B"
    (its-define-hangul	"Qjr"	"$(C;6(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qjs"	"$(C;7(B"	kijuhynbmlopP)
    (its-define-hangul	"Qje"	"$(C;8(B"	kijuhynbmlopP)
    (its-define-hangul	"Qjf"	"$(C;9(B"	kijuhynbmlopP)
    (its-define-hangul	"Qja"	"$(C;:(B"	kijuhynbmlopP)
    (its-define-hangul	"Qjt"	"$(C;;(B"	kijuhynbmlopOP)
    (its-define-hangul	"QjT"	"$(C;<(B"	kjhynmlop)
    (its-define-hangul	"Qjd"	"$(C;=(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qp"	"$(C;>(B"
    (its-define-hangul	"Qpd"	"$(C;?(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qu"	"$(C;@(B"
    (its-define-hangul	"Qur"	"$(C;A(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qua"	"$(C;B(B"	kijuhynbmlopP)
    (its-define-hangul	"Quq"	"$(C;C(B"	kijuhynbmlopP)
    (its-define-hangul	"Qut"	"$(C;D(B"	kijuhynbmlopOP)
    (its-define-hangul	"QuT"	"$(C;E(B"	kjhynmlop)
    (its-define-hangul	"Qud"	"$(C;F(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qh"	"$(C;G(B"
    (its-define-hangul	"Qhr"	"$(C;H(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qhs"	"$(C;I(B"	kijuhynbmlopP)
    (its-define-hangul	"Qhf"	"$(C;J(B"	kijuhynbmlopP)
    (its-define-hangul	"Qha"	"$(C;K(B"	kijuhynbmlopP)
    (its-define-hangul	"Qhq"	"$(C;L(B"	kijuhynbmlopP)
    (its-define-hangul	"Qhd"	"$(C;M(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"Qhl"	"$(C;N(B")

  (its-defrule-hangul-2	"Qy"	"$(C;O(B"
    (its-define-hangul	"Qyd"	"$(C;P(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qn"	"$(C;Q(B"
    (its-define-hangul	"Qnr"	"$(C;R(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qns"	"$(C;S(B"	kijuhynbmlopP)
    (its-define-hangul	"Qnf"	"$(C;T(B"	kijuhynbmlopP)
    (its-define-hangul	"Qna"	"$(C;U(B"	kijuhynbmlopP)
    (its-define-hangul	"Qnt"	"$(C;V(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qnd"	"$(C;W(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qb"	"$(C;X(B"
    (its-define-hangul	"Qbd"	"$(C;Y(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Qm"	"$(C;Z(B"
    (its-define-hangul	"Qms"	"$(C;[(B"	kijuhynbmlopP)
    (its-define-hangul	"Qmf"	"$(C;\(B"	kijuhynbmlopP)
    (its-define-hangul	"Qma"	"$(C;](B"	kijuhynbmlopP)
    (its-define-hangul	"Qmq"	"$(C;^(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Ql"	"$(C;_(B"
    (its-define-hangul	"Qlr"	"$(C;`(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qls"	"$(C;a(B"	kijuhynbmlopP)
    (its-define-hangul	"Qlf"	"$(C;b(B"	kijuhynbmlopP)
    (its-define-hangul	"Qla"	"$(C;c(B"	kijuhynbmlopP)
    (its-define-hangul	"Qlq"	"$(C;d(B"	kijuhynbmlopP)
    (its-define-hangul	"Qlt"	"$(C;e(B"	kijuhynbmlopOP)
    (its-define-hangul	"Qld"	"$(C;f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tk"	"$(C;g(B"
    (its-define-hangul	"tkr"	"$(C;h(B"	kijuhynbmlopOP
     (its-define-hangul	"tkrt"	"$(C;i(B"	kijuhynbmlopOP))
    (its-define-hangul	"tks"	"$(C;j(B"	kijuhynbmlopP)
    (its-define-hangul	"tke"	"$(C;k(B"	kijuhynbmlopP)
    (its-define-hangul	"tkf"	"$(C;l(B"	kijuhynbmlopP
     (its-define-hangul	"tkfr"	"$(C;m(B"	kijuhynbmlopOP)
     (its-define-hangul	"tkfa"	"$(C;n(B"	kijuhynbmlopP))
    (its-define-hangul	"tka"	"$(C;o(B"	kijuhynbmlopP)
    (its-define-hangul	"tkq"	"$(C;p(B"	kijuhynbmlopP)
    (its-define-hangul	"tkt"	"$(C;q(B"	kijuhynbmlopOP
     (its-define-hangul	"tktT"	"$(C;r(B"	kjhynmlop))
    (its-define-hangul	"tkd"	"$(C;s(B"	kijuhynbmlopOP)
    (its-define-hangul	"tkx"	"$(C;t(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"to"	"$(C;u(B"
    (its-define-hangul	"tor"	"$(C;v(B"	kijuhynbmlopOP)
    (its-define-hangul	"tos"	"$(C;w(B"	kijuhynbmlopP)
    (its-define-hangul	"tof"	"$(C;x(B"	kijuhynbmlopP)
    (its-define-hangul	"toa"	"$(C;y(B"	kijuhynbmlopP)
    (its-define-hangul	"toq"	"$(C;z(B"	kijuhynbmlopP)
    (its-define-hangul	"tot"	"$(C;{(B"	kijuhynbmlopOP)
    (its-define-hangul	"toT"	"$(C;|(B"	kjhynmlop)
    (its-define-hangul	"tod"	"$(C;}(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ti"	"$(C;~(B"
    (its-define-hangul	"tir"	"$(C<!(B"	kijuhynbmlopOP)
    (its-define-hangul	"tis"	"$(C<"(B"	kijuhynbmlopP)
    (its-define-hangul	"tif"	"$(C<#(B"	kijuhynbmlopP)
    (its-define-hangul	"tia"	"$(C<$(B"	kijuhynbmlopP)
    (its-define-hangul	"tiq"	"$(C<%(B"	kijuhynbmlopP)
    (its-define-hangul	"tit"	"$(C<&(B"	kijuhynbmlopOP)
    (its-define-hangul	"tid"	"$(C<'(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tO"	"$(C<((B"
    (its-define-hangul	"tOs"	"$(C<)(B"	kijuhynbmlopP)
    (its-define-hangul	"tOf"	"$(C<*(B"	kijuhynbmlopP)
    (its-define-hangul	"tOa"	"$(C<+(B"	kijuhynbmlopP)
    (its-define-hangul	"tOd"	"$(C<,(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tj"	"$(C<-(B"
    (its-define-hangul	"tjr"	"$(C<.(B"	kijuhynbmlopOP
     (its-define-hangul	"tjrt"	"$(C<0(B"	kijuhynbmlopOP))
    (its-define-hangul	"tjR"	"$(C</(B"	kijuhynbmlopP)
    (its-define-hangul	"tjs"	"$(C<1(B"	kijuhynbmlopP)
    (its-define-hangul	"tje"	"$(C<2(B"	kijuhynbmlopP)
    (its-define-hangul	"tjf"	"$(C<3(B"	kijuhynbmlopP
     (its-define-hangul	"tjfa"	"$(C<4(B"	kijuhynbmlopP)
     (its-define-hangul	"tjfq"	"$(C<5(B"	kijuhynbmlopP))
    (its-define-hangul	"tja"	"$(C<6(B"	kijuhynbmlopP)
    (its-define-hangul	"tjq"	"$(C<7(B"	kijuhynbmlopP)
    (its-define-hangul	"tjt"	"$(C<8(B"	kijuhynbmlopOP)
    (its-define-hangul	"tjT"	"$(C<9(B"	kjhynmlop)
    (its-define-hangul	"tjd"	"$(C<:(B"	kijuhynbmlopOP)
    (its-define-hangul	"tjv"	"$(C<;(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"tp"	"$(C<<(B"
    (its-define-hangul	"tpr"	"$(C<=(B"	kijuhynbmlopOP)
    (its-define-hangul	"tps"	"$(C<>(B"	kijuhynbmlopP)
    (its-define-hangul	"tpf"	"$(C<?(B"	kijuhynbmlopP)
    (its-define-hangul	"tpa"	"$(C<@(B"	kijuhynbmlopP)
    (its-define-hangul	"tpq"	"$(C<A(B"	kijuhynbmlopP)
    (its-define-hangul	"tpt"	"$(C<B(B"	kijuhynbmlopOP)
    (its-define-hangul	"tpT"	"$(C<C(B"	kjhynmlop)
    (its-define-hangul	"tpd"	"$(C<D(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tu"	"$(C<E(B"
    (its-define-hangul	"tur"	"$(C<F(B"	kijuhynbmlopOP)
    (its-define-hangul	"tus"	"$(C<G(B"	kijuhynbmlopP)
    (its-define-hangul	"tuf"	"$(C<H(B"	kijuhynbmlopP)
    (its-define-hangul	"tua"	"$(C<I(B"	kijuhynbmlopP)
    (its-define-hangul	"tuq"	"$(C<J(B"	kijuhynbmlopP)
    (its-define-hangul	"tut"	"$(C<K(B"	kijuhynbmlopOP)
    (its-define-hangul	"tuT"	"$(C<L(B"	kjhynmlop)
    (its-define-hangul	"tud"	"$(C<M(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tP"	"$(C<N(B"
    (its-define-hangul	"tPs"	"$(C<O(B"	kijuhynbmlopP)
    (its-define-hangul	"tPf"	"$(C<P(B"	kijuhynbmlopP)
    (its-define-hangul	"tPd"	"$(C<Q(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"th"	"$(C<R(B"
    (its-define-hangul	"thr"	"$(C<S(B"	kijuhynbmlopOP)
    (its-define-hangul	"thR"	"$(C<T(B"	kijuhynbmlopP)
    (its-define-hangul	"ths"	"$(C<U(B"	kijuhynbmlopP)
    (its-define-hangul	"thf"	"$(C<V(B"	kijuhynbmlopP
     (its-define-hangul	"thfa"	"$(C<W(B"	kijuhynbmlopP))
    (its-define-hangul	"tha"	"$(C<X(B"	kijuhynbmlopP)
    (its-define-hangul	"thq"	"$(C<Y(B"	kijuhynbmlopP)
    (its-define-hangul	"tht"	"$(C<Z(B"	kijuhynbmlopOP)
    (its-define-hangul	"thd"	"$(C<[(B"	kijuhynbmlopOP)
    (its-define-hangul	"thx"	"$(C<\(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"thk"	"$(C<](B"
    (its-define-hangul	"thkr"	"$(C<^(B"	kijuhynbmlopOP)
    (its-define-hangul	"thks"	"$(C<_(B"	kijuhynbmlopP)
    (its-define-hangul	"thkf"	"$(C<`(B"	kijuhynbmlopP)
    (its-define-hangul	"thkd"	"$(C<a(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tho"	"$(C<b(B"
    (its-define-hangul	"thos"	"$(C<c(B"	kijuhynbmlopP)
    (its-define-hangul	"thof"	"$(C<d(B"	kijuhynbmlopP)
    (its-define-hangul	"thoa"	"$(C<e(B"	kijuhynbmlopP)
    (its-define-hangul	"thot"	"$(C<f(B"	kijuhynbmlopOP)
    (its-define-hangul	"thoT"	"$(C<g(B"	kjhynmlop))

  (its-defrule-hangul-2	"thl"	"$(C<h(B"
    (its-define-hangul	"thls"	"$(C<i(B"	kijuhynbmlopP)
    (its-define-hangul	"thlf"	"$(C<j(B"	kijuhynbmlopP)
    (its-define-hangul	"thla"	"$(C<k(B"	kijuhynbmlopP)
    (its-define-hangul	"thlq"	"$(C<l(B"	kijuhynbmlopP)
    (its-define-hangul	"thlt"	"$(C<m(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ty"	"$(C<n(B"
    (its-define-hangul	"tyr"	"$(C<o(B"	kijuhynbmlopOP)
    (its-define-hangul	"tys"	"$(C<p(B"	kijuhynbmlopP)
    (its-define-hangul	"tyf"	"$(C<q(B"	kijuhynbmlopP)
    (its-define-hangul	"tya"	"$(C<r(B"	kijuhynbmlopP)
    (its-define-hangul	"tyq"	"$(C<s(B"	kijuhynbmlopP)
    (its-define-hangul	"tyt"	"$(C<t(B"	kijuhynbmlopOP)
    (its-define-hangul	"tyd"	"$(C<u(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tn"	"$(C<v(B"
    (its-define-hangul	"tnr"	"$(C<w(B"	kijuhynbmlopOP)
    (its-define-hangul	"tns"	"$(C<x(B"	kijuhynbmlopP)
    (its-define-hangul	"tne"	"$(C<y(B"	kijuhynbmlopP)
    (its-define-hangul	"tnf"	"$(C<z(B"	kijuhynbmlopP)
    (its-define-hangul	"tna"	"$(C<{(B"	kijuhynbmlopP)
    (its-define-hangul	"tnq"	"$(C<|(B"	kijuhynbmlopP)
    (its-define-hangul	"tnt"	"$(C<}(B"	kijuhynbmlopOP)
    (its-define-hangul	"tnd"	"$(C<~(B"	kijuhynbmlopOP)
    (its-define-hangul	"tnc"	"$(C=!(B"	kijuhynbmlopP)
    (its-define-hangul	"tnx"	"$(C="(B"	kijuhynbmlopP)
    (its-define-hangul	"tnv"	"$(C=#(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"tnj"	"$(C=$(B"
    (its-define-hangul	"tnjT"	"$(C=%(B"	kjhynmlop))

  (its-defrule-hangul-2	"tnp"	"$(C=&(B"
    (its-define-hangul	"tnpr"	"$(C='(B"	kijuhynbmlopOP)
    (its-define-hangul	"tnps"	"$(C=((B"	kijuhynbmlopP)
    (its-define-hangul	"tnpf"	"$(C=)(B"	kijuhynbmlopP)
    (its-define-hangul	"tnpa"	"$(C=*(B"	kijuhynbmlopP)
    (its-define-hangul	"tnpd"	"$(C=+(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tnl"	"$(C=,(B"
    (its-define-hangul	"tnlr"	"$(C=-(B"	kijuhynbmlopOP)
    (its-define-hangul	"tnls"	"$(C=.(B"	kijuhynbmlopP)
    (its-define-hangul	"tnlf"	"$(C=/(B"	kijuhynbmlopP)
    (its-define-hangul	"tnla"	"$(C=0(B"	kijuhynbmlopP)
    (its-define-hangul	"tnlq"	"$(C=1(B"	kijuhynbmlopP)
    (its-define-hangul	"tnlt"	"$(C=2(B"	kijuhynbmlopOP)
    (its-define-hangul	"tnld"	"$(C=3(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tb"	"$(C=4(B"
    (its-define-hangul	"tbr"	"$(C=5(B"	kijuhynbmlopOP)
    (its-define-hangul	"tbf"	"$(C=6(B"	kijuhynbmlopP)
    (its-define-hangul	"tba"	"$(C=7(B"	kijuhynbmlopP)
    (its-define-hangul	"tbt"	"$(C=8(B"	kijuhynbmlopOP)
    (its-define-hangul	"tbd"	"$(C=9(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tm"	"$(C=:(B"
    (its-define-hangul	"tmr"	"$(C=;(B"	kijuhynbmlopOP)
    (its-define-hangul	"tms"	"$(C=<(B"	kijuhynbmlopP)
    (its-define-hangul	"tmf"	"$(C==(B"	kijuhynbmlopP
     (its-define-hangul	"tmfr"	"$(C=>(B"	kijuhynbmlopOP))
    (its-define-hangul	"tma"	"$(C=?(B"	kijuhynbmlopP)
    (its-define-hangul	"tmq"	"$(C=@(B"	kijuhynbmlopP)
    (its-define-hangul	"tmt"	"$(C=A(B"	kijuhynbmlopOP)
    (its-define-hangul	"tmd"	"$(C=B(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"tl"	"$(C=C(B"
    (its-define-hangul	"tlr"	"$(C=D(B"	kijuhynbmlopOP)
    (its-define-hangul	"tls"	"$(C=E(B"	kijuhynbmlopP)
    (its-define-hangul	"tle"	"$(C=F(B"	kijuhynbmlopP)
    (its-define-hangul	"tlf"	"$(C=G(B"	kijuhynbmlopP
     (its-define-hangul	"tlfg"	"$(C=H(B"	kijuhynbmlopP))
    (its-define-hangul	"tla"	"$(C=I(B"	kijuhynbmlopP)
    (its-define-hangul	"tlq"	"$(C=J(B"	kijuhynbmlopP)
    (its-define-hangul	"tlt"	"$(C=K(B"	kijuhynbmlopOP)
    (its-define-hangul	"tld"	"$(C=L(B"	kijuhynbmlopOP)
    (its-define-hangul	"tlv"	"$(C=M(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Tk"	"$(C=N(B"
    (its-define-hangul	"Tkr"	"$(C=O(B"	kijuhynbmlopOP
     (its-define-hangul	"Tkrt"	"$(C=P(B"	kijuhynbmlopOP))
    (its-define-hangul	"Tks"	"$(C=Q(B"	kijuhynbmlopP)
    (its-define-hangul	"Tkf"	"$(C=R(B"	kijuhynbmlopP)
    (its-define-hangul	"Tka"	"$(C=S(B"	kijuhynbmlopP)
    (its-define-hangul	"Tkq"	"$(C=T(B"	kijuhynbmlopP)
    (its-define-hangul	"TkT"	"$(C=U(B"	kjhynmlop)
    (its-define-hangul	"Tkd"	"$(C=V(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tkg"	"$(C=W(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"To"	"$(C=X(B"
    (its-define-hangul	"Tor"	"$(C=Y(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tos"	"$(C=Z(B"	kijuhynbmlopP)
    (its-define-hangul	"Tof"	"$(C=[(B"	kijuhynbmlopP)
    (its-define-hangul	"Toa"	"$(C=\(B"	kijuhynbmlopP)
    (its-define-hangul	"Toq"	"$(C=](B"	kijuhynbmlopP)
    (its-define-hangul	"ToT"	"$(C=^(B"	kjhynmlop)
    (its-define-hangul	"Tod"	"$(C=_(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"Ti"	"$(C$6(Bi")
  (its-defrule-hangul	"Tid"	"$(C=`(B")

  (its-defrule-hangul-2	"Tj"	"$(C=a(B"
    (its-define-hangul	"Tjr"	"$(C=b(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tjs"	"$(C=c(B"	kijuhynbmlopP)
    (its-define-hangul	"Tjf"	"$(C=d(B"	kijuhynbmlopP
     (its-define-hangul	"Tjfq"	"$(C=e(B"	kijuhynbmlopP))
    (its-define-hangul	"Tja"	"$(C=f(B"	kijuhynbmlopP)
    (its-define-hangul	"Tjq"	"$(C=g(B"	kijuhynbmlopP)
    (its-define-hangul	"TjT"	"$(C=h(B"	kjhynmlop)
    (its-define-hangul	"Tjd"	"$(C=i(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Tp"	"$(C=j(B"
    (its-define-hangul	"Tps"	"$(C=k(B"	kijuhynbmlopP)
    (its-define-hangul	"Tpf"	"$(C=l(B"	kijuhynbmlopP))

  (its-defrule-hangul	"TP"	"$(C$6(BP")
  (its-defrule-hangul	"TPs"	"$(C=m(B")

  (its-defrule-hangul-2	"Th"	"$(C=n(B"
    (its-define-hangul	"Thr"	"$(C=o(B"	kijuhynbmlopOP)
    (its-define-hangul	"Ths"	"$(C=p(B"	kijuhynbmlopP)
    (its-define-hangul	"The"	"$(C=q(B"	kijuhynbmlopP)
    (its-define-hangul	"Thf"	"$(C=r(B"	kijuhynbmlopP
     (its-define-hangul	"Thfa"	"$(C=s(B"	kijuhynbmlopP))
    (its-define-hangul	"Tha"	"$(C=t(B"	kijuhynbmlopP)
    (its-define-hangul	"Thq"	"$(C=u(B"	kijuhynbmlopP)
    (its-define-hangul	"Thd"	"$(C=v(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Thk"	"$(C=w(B"
    (its-define-hangul	"Thkr"	"$(C=x(B"	kijuhynbmlopOP)
    (its-define-hangul	"Thks"	"$(C=y(B"	kijuhynbmlopP)
    (its-define-hangul	"ThkT"	"$(C=z(B"	kjhynmlop))

  (its-defrule-hangul-2	"Tho"	"$(C={(B"
    (its-define-hangul	"ThoT"	"$(C=|(B"	kjhynmlop))

  (its-defrule-hangul-2	"Thl"	"$(C=}(B"
    (its-define-hangul	"Thls"	"$(C=~(B"	kijuhynbmlopP)
    (its-define-hangul	"Thlf"	"$(C>!(B"	kijuhynbmlopP)
    (its-define-hangul	"Thla"	"$(C>"(B"	kijuhynbmlopP)
    (its-define-hangul	"Thlq"	"$(C>#(B"	kijuhynbmlopP))

  (its-defrule-hangul	"Ty"	"$(C>$(B")

  (its-defrule-hangul-2	"Tn"	"$(C>%(B"
    (its-define-hangul	"Tnr"	"$(C>&(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tns"	"$(C>'(B"	kijuhynbmlopP)
    (its-define-hangul	"Tnf"	"$(C>((B"	kijuhynbmlopP)
    (its-define-hangul	"Tna"	"$(C>)(B"	kijuhynbmlopP)
    (its-define-hangul	"Tnq"	"$(C>*(B"	kijuhynbmlopP)
    (its-define-hangul	"Tnd"	"$(C>+(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Tnj"	"$(C>,(B"
    (its-define-hangul	"TnjT"	"$(C>-(B"	kjhynmlop))

  (its-defrule-hangul	"Tnjp"	"$(C>.(B")

  (its-defrule-hangul-2	"Tnl"	"$(C>/(B"
    (its-define-hangul	"Tnls"	"$(C>0(B"	kijuhynbmlopP))

  (its-defrule-hangul	"Tb"	"$(C$6(Bd")
  (its-defrule-hangul	"Tbd"	"$(C>1(B")

  (its-defrule-hangul-2	"Tm"	"$(C>2(B"
    (its-define-hangul	"Tmr"	"$(C>3(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tms"	"$(C>4(B"	kijuhynbmlopP)
    (its-define-hangul	"Tmf"	"$(C>5(B"	kijuhynbmlopP
     (its-define-hangul	"Tmfa"	"$(C>6(B"	kijuhynbmlopP)
     (its-define-hangul	"Tmfg"	"$(C>7(B"	kijuhynbmlopP))
    (its-define-hangul	"Tma"	"$(C>8(B"	kijuhynbmlopP)
    (its-define-hangul	"Tmq"	"$(C>9(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Tml"	"$(C>:(B"
    (its-define-hangul	"Tmls"	"$(C>;(B"	kijuhynbmlopP)
    (its-define-hangul	"Tmlf"	"$(C><(B"	kijuhynbmlopP)
    (its-define-hangul	"Tmla"	"$(C>=(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Tl"	"$(C>>(B"
    (its-define-hangul	"Tlr"	"$(C>?(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tls"	"$(C>@(B"	kijuhynbmlopP)
    (its-define-hangul	"Tlf"	"$(C>A(B"	kijuhynbmlopP)
    (its-define-hangul	"Tla"	"$(C>B(B"	kijuhynbmlopP)
    (its-define-hangul	"Tlq"	"$(C>C(B"	kijuhynbmlopP)
    (its-define-hangul	"Tlt"	"$(C>D(B"	kijuhynbmlopOP)
    (its-define-hangul	"Tld"	"$(C>E(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dk"	"$(C>F(B"
    (its-define-hangul	"dkr"	"$(C>G(B"	kijuhynbmlopOP)
    (its-define-hangul	"dks"	"$(C>H(B"	kijuhynbmlopP
     (its-define-hangul	"dksw"	"$(C>I(B"	kijuhynbmlopOP)
     (its-define-hangul	"dksg"	"$(C>J(B"	kijuhynbmlopP))
    (its-define-hangul	"dkf"	"$(C>K(B"	kijuhynbmlopP
     (its-define-hangul	"dkfr"	"$(C>L(B"	kijuhynbmlopOP)
     (its-define-hangul	"dkfa"	"$(C>M(B"	kijuhynbmlopP)
     (its-define-hangul	"dkfg"	"$(C>N(B"	kijuhynbmlopP))
    (its-define-hangul	"dka"	"$(C>O(B"	kijuhynbmlopP)
    (its-define-hangul	"dkq"	"$(C>P(B"	kijuhynbmlopP)
    (its-define-hangul	"dkt"	"$(C>Q(B"	kijuhynbmlopOP)
    (its-define-hangul	"dkT"	"$(C>R(B"	kjhynmlop)
    (its-define-hangul	"dkd"	"$(C>S(B"	kijuhynbmlopOP)
    (its-define-hangul	"dkx"	"$(C>T(B"	kijuhynbmlopP)
    (its-define-hangul	"dkv"	"$(C>U(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"do"	"$(C>V(B"
    (its-define-hangul	"dor"	"$(C>W(B"	kijuhynbmlopOP)
    (its-define-hangul	"dos"	"$(C>X(B"	kijuhynbmlopP)
    (its-define-hangul	"dof"	"$(C>Y(B"	kijuhynbmlopP)
    (its-define-hangul	"doa"	"$(C>Z(B"	kijuhynbmlopP)
    (its-define-hangul	"doq"	"$(C>[(B"	kijuhynbmlopP)
    (its-define-hangul	"dot"	"$(C>\(B"	kijuhynbmlopOP)
    (its-define-hangul	"doT"	"$(C>](B"	kjhynmlop)
    (its-define-hangul	"dod"	"$(C>^(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"di"	"$(C>_(B"
    (its-define-hangul	"dir"	"$(C>`(B"	kijuhynbmlopOP)
    (its-define-hangul	"dis"	"$(C>a(B"	kijuhynbmlopP)
    (its-define-hangul	"dif"	"$(C>b(B"	kijuhynbmlopP
     (its-define-hangul	"difq"	"$(C>c(B"	kijuhynbmlopP))
    (its-define-hangul	"dia"	"$(C>d(B"	kijuhynbmlopP)
    (its-define-hangul	"diq"	"$(C>e(B"	kijuhynbmlopP)
    (its-define-hangul	"dit"	"$(C>f(B"	kijuhynbmlopOP)
    (its-define-hangul	"did"	"$(C>g(B"	kijuhynbmlopOP)
    (its-define-hangul	"dix"	"$(C>h(B"	kijuhynbmlopP)
    (its-define-hangul	"dig"	"$(C>i(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dO"	"$(C>j(B"
    (its-define-hangul	"dOs"	"$(C>k(B"	kijuhynbmlopP)
    (its-define-hangul	"dOf"	"$(C>l(B"	kijuhynbmlopP)
    (its-define-hangul	"dOq"	"$(C>m(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dj"	"$(C>n(B"
    (its-define-hangul	"djr"	"$(C>o(B"	kijuhynbmlopOP)
    (its-define-hangul	"djs"	"$(C>p(B"	kijuhynbmlopP
     (its-define-hangul	"djsw"	"$(C>q(B"	kijuhynbmlopOP))
    (its-define-hangul	"dje"	"$(C>r(B"	kijuhynbmlopP)
    (its-define-hangul	"djf"	"$(C>s(B"	kijuhynbmlopP
     (its-define-hangul	"djfr"	"$(C>t(B"	kijuhynbmlopOP)
     (its-define-hangul	"djfa"	"$(C>u(B"	kijuhynbmlopP))
    (its-define-hangul	"dja"	"$(C>v(B"	kijuhynbmlopP)
    (its-define-hangul	"djq"	"$(C>w(B"	kijuhynbmlopP
     (its-define-hangul	"djqt"	"$(C>x(B"	kijuhynbmlopOP))
    (its-define-hangul	"djt"	"$(C>y(B"	kijuhynbmlopOP)
    (its-define-hangul	"djT"	"$(C>z(B"	kjhynmlop)
    (its-define-hangul	"djd"	"$(C>{(B"	kijuhynbmlopOP)
    (its-define-hangul	"djw"	"$(C>|(B"	kijuhynbmlopOP)
    (its-define-hangul	"djz"	"$(C>}(B"	kijuhynbmlopP)
    (its-define-hangul	"djv"	"$(C>~(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dp"	"$(C?!(B"
    (its-define-hangul	"dpr"	"$(C?"(B"	kijuhynbmlopOP)
    (its-define-hangul	"dps"	"$(C?#(B"	kijuhynbmlopP)
    (its-define-hangul	"dpf"	"$(C?$(B"	kijuhynbmlopP)
    (its-define-hangul	"dpa"	"$(C?%(B"	kijuhynbmlopP)
    (its-define-hangul	"dpq"	"$(C?&(B"	kijuhynbmlopP)
    (its-define-hangul	"dpt"	"$(C?'(B"	kijuhynbmlopOP)
    (its-define-hangul	"dpd"	"$(C?((B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"du"	"$(C?)(B"
    (its-define-hangul	"dur"	"$(C?*(B"	kijuhynbmlopOP)
    (its-define-hangul	"duR"	"$(C?+(B"	kijuhynbmlopP)
    (its-define-hangul	"dus"	"$(C?,(B"	kijuhynbmlopP)
    (its-define-hangul	"duf"	"$(C?-(B"	kijuhynbmlopP
     (its-define-hangul	"dufa"	"$(C?.(B"	kijuhynbmlopP)
     (its-define-hangul	"dufq"	"$(C?/(B"	kijuhynbmlopP))
    (its-define-hangul	"dua"	"$(C?0(B"	kijuhynbmlopP)
    (its-define-hangul	"duq"	"$(C?1(B"	kijuhynbmlopP
     (its-define-hangul	"duqt"	"$(C?2(B"	kijuhynbmlopOP))
    (its-define-hangul	"dut"	"$(C?3(B"	kijuhynbmlopOP)
    (its-define-hangul	"duT"	"$(C?4(B"	kjhynmlop)
    (its-define-hangul	"dud"	"$(C?5(B"	kijuhynbmlopOP)
    (its-define-hangul	"dux"	"$(C?6(B"	kijuhynbmlopP)
    (its-define-hangul	"duv"	"$(C?7(B"	kijuhynbmlopP)
    (its-define-hangul	"dug"	"$(C?8(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dP"	"$(C?9(B"
    (its-define-hangul	"dPs"	"$(C?:(B"	kijuhynbmlopP)
    (its-define-hangul	"dPf"	"$(C?;(B"	kijuhynbmlopP)
    (its-define-hangul	"dPa"	"$(C?<(B"	kijuhynbmlopP)
    (its-define-hangul	"dPq"	"$(C?=(B"	kijuhynbmlopP)
    (its-define-hangul	"dPt"	"$(C?>(B"	kijuhynbmlopOP)
    (its-define-hangul	"dPT"	"$(C??(B"	kjhynmlop))

  (its-defrule-hangul-2	"dh"	"$(C?@(B"
    (its-define-hangul	"dhr"	"$(C?A(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhs"	"$(C?B(B"	kijuhynbmlopP)
    (its-define-hangul	"dhf"	"$(C?C(B"	kijuhynbmlopP
     (its-define-hangul	"dhfr"	"$(C?D(B"	kijuhynbmlopOP)
     (its-define-hangul	"dhfa"	"$(C?E(B"	kijuhynbmlopP)
     (its-define-hangul	"dhft"	"$(C?F(B"	kijuhynbmlopOP)
     (its-define-hangul	"dhfg"	"$(C?G(B"	kijuhynbmlopP))
    (its-define-hangul	"dha"	"$(C?H(B"	kijuhynbmlopP)
    (its-define-hangul	"dhq"	"$(C?I(B"	kijuhynbmlopP)
    (its-define-hangul	"dht"	"$(C?J(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhd"	"$(C?K(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhc"	"$(C?L(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dhk"	"$(C?M(B"
    (its-define-hangul	"dhkr"	"$(C?N(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhks"	"$(C?O(B"	kijuhynbmlopP)
    (its-define-hangul	"dhkf"	"$(C?P(B"	kijuhynbmlopP)
    (its-define-hangul	"dhka"	"$(C?Q(B"	kijuhynbmlopP)
    (its-define-hangul	"dhkq"	"$(C?R(B"	kijuhynbmlopP)
    (its-define-hangul	"dhkt"	"$(C?S(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhkT"	"$(C?T(B"	kjhynmlop)
    (its-define-hangul	"dhkd"	"$(C?U(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dho"	"$(C?V(B"
    (its-define-hangul	"dhor"	"$(C?W(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhos"	"$(C?X(B"	kijuhynbmlopP)
    (its-define-hangul	"dhoa"	"$(C?Y(B"	kijuhynbmlopP)
    (its-define-hangul	"dhot"	"$(C?Z(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhod"	"$(C?[(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dhl"	"$(C?\(B"
    (its-define-hangul	"dhlr"	"$(C?](B"	kijuhynbmlopOP)
    (its-define-hangul	"dhls"	"$(C?^(B"	kijuhynbmlopP)
    (its-define-hangul	"dhlf"	"$(C?_(B"	kijuhynbmlopP)
    (its-define-hangul	"dhla"	"$(C?`(B"	kijuhynbmlopP)
    (its-define-hangul	"dhlq"	"$(C?a(B"	kijuhynbmlopP)
    (its-define-hangul	"dhlt"	"$(C?b(B"	kijuhynbmlopOP)
    (its-define-hangul	"dhld"	"$(C?c(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dy"	"$(C?d(B"
    (its-define-hangul	"dyr"	"$(C?e(B"	kijuhynbmlopOP)
    (its-define-hangul	"dys"	"$(C?f(B"	kijuhynbmlopP)
    (its-define-hangul	"dyf"	"$(C?g(B"	kijuhynbmlopP)
    (its-define-hangul	"dya"	"$(C?h(B"	kijuhynbmlopP)
    (its-define-hangul	"dyq"	"$(C?i(B"	kijuhynbmlopP)
    (its-define-hangul	"dyt"	"$(C?j(B"	kijuhynbmlopOP)
    (its-define-hangul	"dyd"	"$(C?k(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dn"	"$(C?l(B"
    (its-define-hangul	"dnr"	"$(C?m(B"	kijuhynbmlopOP)
    (its-define-hangul	"dns"	"$(C?n(B"	kijuhynbmlopP)
    (its-define-hangul	"dnf"	"$(C?o(B"	kijuhynbmlopP
     (its-define-hangul	"dnfr"	"$(C?p(B"	kijuhynbmlopOP)
     (its-define-hangul	"dnfa"	"$(C?q(B"	kijuhynbmlopP))
    (its-define-hangul	"dna"	"$(C?r(B"	kijuhynbmlopP)
    (its-define-hangul	"dnq"	"$(C?s(B"	kijuhynbmlopP)
    (its-define-hangul	"dnt"	"$(C?t(B"	kijuhynbmlopOP)
    (its-define-hangul	"dnd"	"$(C?u(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dnj"	"$(C?v(B"
    (its-define-hangul	"dnjr"	"$(C?w(B"	kijuhynbmlopOP)
    (its-define-hangul	"dnjs"	"$(C?x(B"	kijuhynbmlopP)
    (its-define-hangul	"dnjf"	"$(C?y(B"	kijuhynbmlopP)
    (its-define-hangul	"dnja"	"$(C?z(B"	kijuhynbmlopP)
    (its-define-hangul	"dnjq"	"$(C?{(B"	kijuhynbmlopP)
    (its-define-hangul	"dnjT"	"$(C?|(B"	kjhynmlop)
    (its-define-hangul	"dnjd"	"$(C?}(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dnp"	"$(C?~(B"
    (its-define-hangul	"dnpr"	"$(C@!(B"	kijuhynbmlopOP)
    (its-define-hangul	"dnps"	"$(C@"(B"	kijuhynbmlopP)
    (its-define-hangul	"dnpf"	"$(C@#(B"	kijuhynbmlopP)
    (its-define-hangul	"dnpa"	"$(C@$(B"	kijuhynbmlopP)
    (its-define-hangul	"dnpq"	"$(C@%(B"	kijuhynbmlopP)
    (its-define-hangul	"dnpd"	"$(C@&(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dnl"	"$(C@'(B"
    (its-define-hangul	"dnlr"	"$(C@((B"	kijuhynbmlopOP)
    (its-define-hangul	"dnls"	"$(C@)(B"	kijuhynbmlopP)
    (its-define-hangul	"dnlf"	"$(C@*(B"	kijuhynbmlopP)
    (its-define-hangul	"dnla"	"$(C@+(B"	kijuhynbmlopP)
    (its-define-hangul	"dnlq"	"$(C@,(B"	kijuhynbmlopP)
    (its-define-hangul	"dnlt"	"$(C@-(B"	kijuhynbmlopOP)
    (its-define-hangul	"dnld"	"$(C@.(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"db"	"$(C@/(B"
    (its-define-hangul	"dbr"	"$(C@0(B"	kijuhynbmlopOP)
    (its-define-hangul	"dbs"	"$(C@1(B"	kijuhynbmlopP)
    (its-define-hangul	"dbf"	"$(C@2(B"	kijuhynbmlopP)
    (its-define-hangul	"dba"	"$(C@3(B"	kijuhynbmlopP)
    (its-define-hangul	"dbq"	"$(C@4(B"	kijuhynbmlopP)
    (its-define-hangul	"dbt"	"$(C@5(B"	kijuhynbmlopOP)
    (its-define-hangul	"dbd"	"$(C@6(B"	kijuhynbmlopOP)
    (its-define-hangul	"dbc"	"$(C@7(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dm"	"$(C@8(B"
    (its-define-hangul	"dmr"	"$(C@9(B"	kijuhynbmlopOP)
    (its-define-hangul	"dms"	"$(C@:(B"	kijuhynbmlopP)
    (its-define-hangul	"dmf"	"$(C@;(B"	kijuhynbmlopP
     (its-define-hangul	"dmfv"	"$(C@<(B"	kijuhynbmlopP))
    (its-define-hangul	"dma"	"$(C@=(B"	kijuhynbmlopP)
    (its-define-hangul	"dmq"	"$(C@>(B"	kijuhynbmlopP)
    (its-define-hangul	"dmt"	"$(C@?(B"	kijuhynbmlopOP)
    (its-define-hangul	"dmd"	"$(C@@(B"	kijuhynbmlopOP)
    (its-define-hangul	"dmw"	"$(C@A(B"	kijuhynbmlopOP)
    (its-define-hangul	"dmc"	"$(C@B(B"	kijuhynbmlopP)
    (its-define-hangul	"dmz"	"$(C@C(B"	kijuhynbmlopP)
    (its-define-hangul	"dmx"	"$(C@D(B"	kijuhynbmlopP)
    (its-define-hangul	"dmv"	"$(C@E(B"	kijuhynbmlopP)
    (its-define-hangul	"dmg"	"$(C@F(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"dml"	"$(C@G(B"
    (its-define-hangul	"dmls"	"$(C@H(B"	kijuhynbmlopP)
    (its-define-hangul	"dmlf"	"$(C@I(B"	kijuhynbmlopP)
    (its-define-hangul	"dmla"	"$(C@J(B"	kijuhynbmlopP)
    (its-define-hangul	"dmlt"	"$(C@K(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"dl"	"$(C@L(B"
    (its-define-hangul	"dlr"	"$(C@M(B"	kijuhynbmlopOP)
    (its-define-hangul	"dls"	"$(C@N(B"	kijuhynbmlopP)
    (its-define-hangul	"dlf"	"$(C@O(B"	kijuhynbmlopP
     (its-define-hangul	"dlfr"	"$(C@P(B"	kijuhynbmlopOP)
     (its-define-hangul	"dlfa"	"$(C@Q(B"	kijuhynbmlopP)
     (its-define-hangul	"dlfg"	"$(C@R(B"	kijuhynbmlopP))
    (its-define-hangul	"dla"	"$(C@S(B"	kijuhynbmlopP)
    (its-define-hangul	"dlq"	"$(C@T(B"	kijuhynbmlopP)
    (its-define-hangul	"dlt"	"$(C@U(B"	kijuhynbmlopOP)
    (its-define-hangul	"dlT"	"$(C@V(B"	kjhynmlop)
    (its-define-hangul	"dld"	"$(C@W(B"	kijuhynbmlopOP)
    (its-define-hangul	"dlw"	"$(C@X(B"	kijuhynbmlopOP)
    (its-define-hangul	"dlv"	"$(C@Y(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"wk"	"$(C@Z(B"
    (its-define-hangul	"wkr"	"$(C@[(B"	kijuhynbmlopOP)
    (its-define-hangul	"wks"	"$(C@\(B"	kijuhynbmlopP
     (its-define-hangul	"wksg"	"$(C@](B"	kijuhynbmlopP))
    (its-define-hangul	"wke"	"$(C@^(B"	kijuhynbmlopP)
    (its-define-hangul	"wkf"	"$(C@_(B"	kijuhynbmlopP
     (its-define-hangul	"wkfa"	"$(C@`(B"	kijuhynbmlopP))
    (its-define-hangul	"wka"	"$(C@a(B"	kijuhynbmlopP)
    (its-define-hangul	"wkq"	"$(C@b(B"	kijuhynbmlopP)
    (its-define-hangul	"wkt"	"$(C@c(B"	kijuhynbmlopOP)
    (its-define-hangul	"wkT"	"$(C@d(B"	kjhynmlop)
    (its-define-hangul	"wkd"	"$(C@e(B"	kijuhynbmlopOP)
    (its-define-hangul	"wkw"	"$(C@f(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wo"	"$(C@g(B"
    (its-define-hangul	"wor"	"$(C@h(B"	kijuhynbmlopOP)
    (its-define-hangul	"wos"	"$(C@i(B"	kijuhynbmlopP)
    (its-define-hangul	"wof"	"$(C@j(B"	kijuhynbmlopP)
    (its-define-hangul	"woa"	"$(C@k(B"	kijuhynbmlopP)
    (its-define-hangul	"woq"	"$(C@l(B"	kijuhynbmlopP)
    (its-define-hangul	"wot"	"$(C@m(B"	kijuhynbmlopOP)
    (its-define-hangul	"woT"	"$(C@n(B"	kjhynmlop)
    (its-define-hangul	"wod"	"$(C@o(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wi"	"$(C@p(B"
    (its-define-hangul	"wir"	"$(C@q(B"	kijuhynbmlopOP)
    (its-define-hangul	"wis"	"$(C@r(B"	kijuhynbmlopP
     (its-define-hangul	"wisg"	"$(C@s(B"	kijuhynbmlopP))
    (its-define-hangul	"wif"	"$(C@t(B"	kijuhynbmlopP)
    (its-define-hangul	"wia"	"$(C@u(B"	kijuhynbmlopP)
    (its-define-hangul	"wid"	"$(C@v(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wO"	"$(C@w(B"
    (its-define-hangul	"wOs"	"$(C@x(B"	kijuhynbmlopP)
    (its-define-hangul	"wOf"	"$(C@y(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"wj"	"$(C@z(B"
    (its-define-hangul	"wjr"	"$(C@{(B"	kijuhynbmlopOP)
    (its-define-hangul	"wjs"	"$(C@|(B"	kijuhynbmlopP)
    (its-define-hangul	"wjf"	"$(C@}(B"	kijuhynbmlopP
     (its-define-hangul	"wjfa"	"$(C@~(B"	kijuhynbmlopP))
    (its-define-hangul	"wja"	"$(CA!(B"	kijuhynbmlopP)
    (its-define-hangul	"wjq"	"$(CA"(B"	kijuhynbmlopP)
    (its-define-hangul	"wjt"	"$(CA#(B"	kijuhynbmlopOP)
    (its-define-hangul	"wjd"	"$(CA$(B"	kijuhynbmlopOP)
    (its-define-hangul	"wjw"	"$(CA%(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wp"	"$(CA&(B"
    (its-define-hangul	"wpr"	"$(CA'(B"	kijuhynbmlopOP)
    (its-define-hangul	"wps"	"$(CA((B"	kijuhynbmlopP)
    (its-define-hangul	"wpf"	"$(CA)(B"	kijuhynbmlopP)
    (its-define-hangul	"wpa"	"$(CA*(B"	kijuhynbmlopP)
    (its-define-hangul	"wpq"	"$(CA+(B"	kijuhynbmlopP)
    (its-define-hangul	"wpt"	"$(CA,(B"	kijuhynbmlopOP)
    (its-define-hangul	"wpd"	"$(CA-(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wu"	"$(CA.(B"
    (its-define-hangul	"wus"	"$(CA/(B"	kijuhynbmlopP)
    (its-define-hangul	"wuf"	"$(CA0(B"	kijuhynbmlopP)
    (its-define-hangul	"wua"	"$(CA1(B"	kijuhynbmlopP)
    (its-define-hangul	"wuq"	"$(CA2(B"	kijuhynbmlopP)
    (its-define-hangul	"wuT"	"$(CA3(B"	kjhynmlop)
    (its-define-hangul	"wud"	"$(CA4(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"wP"	"$(CA5(B")

  (its-defrule-hangul-2	"wh"	"$(CA6(B"
    (its-define-hangul	"whr"	"$(CA7(B"	kijuhynbmlopOP)
    (its-define-hangul	"whs"	"$(CA8(B"	kijuhynbmlopP)
    (its-define-hangul	"whf"	"$(CA9(B"	kijuhynbmlopP
     (its-define-hangul	"whfa"	"$(CA:(B"	kijuhynbmlopP))
    (its-define-hangul	"wha"	"$(CA;(B"	kijuhynbmlopP)
    (its-define-hangul	"whq"	"$(CA<(B"	kijuhynbmlopP)
    (its-define-hangul	"wht"	"$(CA=(B"	kijuhynbmlopOP)
    (its-define-hangul	"whd"	"$(CA>(B"	kijuhynbmlopOP)
    (its-define-hangul	"whw"	"$(CA?(B"	kijuhynbmlopOP)
    (its-define-hangul	"whc"	"$(CA@(B"	kijuhynbmlopP)
    (its-define-hangul	"whg"	"$(CAA(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"whk"	"$(CAB(B"
    (its-define-hangul	"whkr"	"$(CAC(B"	kijuhynbmlopOP)
    (its-define-hangul	"whkf"	"$(CAD(B"	kijuhynbmlopP)
    (its-define-hangul	"whkq"	"$(CAE(B"	kijuhynbmlopP)
    (its-define-hangul	"whkt"	"$(CAF(B"	kijuhynbmlopOP)
    (its-define-hangul	"whkd"	"$(CAG(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"who"	"$(CAH(B"
    (its-define-hangul	"whoT"	"$(CAI(B"	kjhynmlop)
    (its-define-hangul	"whod"	"$(CAJ(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"whl"	"$(CAK(B"
    (its-define-hangul	"whls"	"$(CAL(B"	kijuhynbmlopP)
    (its-define-hangul	"whlf"	"$(CAM(B"	kijuhynbmlopP)
    (its-define-hangul	"whla"	"$(CAN(B"	kijuhynbmlopP)
    (its-define-hangul	"whlq"	"$(CAO(B"	kijuhynbmlopP)
    (its-define-hangul	"whlt"	"$(CAP(B"	kijuhynbmlopOP)
    (its-define-hangul	"whld"	"$(CAQ(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wy"	"$(CAR(B"
    (its-define-hangul	"wyr"	"$(CAS(B"	kijuhynbmlopOP)
    (its-define-hangul	"wys"	"$(CAT(B"	kijuhynbmlopP)
    (its-define-hangul	"wyd"	"$(CAU(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wn"	"$(CAV(B"
    (its-define-hangul	"wnr"	"$(CAW(B"	kijuhynbmlopOP)
    (its-define-hangul	"wns"	"$(CAX(B"	kijuhynbmlopP)
    (its-define-hangul	"wnf"	"$(CAY(B"	kijuhynbmlopP
     (its-define-hangul	"wnfr"	"$(CAZ(B"	kijuhynbmlopOP)
     (its-define-hangul	"wnfa"	"$(CA[(B"	kijuhynbmlopP))
    (its-define-hangul	"wna"	"$(CA\(B"	kijuhynbmlopP)
    (its-define-hangul	"wnq"	"$(CA](B"	kijuhynbmlopP)
    (its-define-hangul	"wnt"	"$(CA^(B"	kijuhynbmlopOP)
    (its-define-hangul	"wnd"	"$(CA_(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wnj"	"$(CA`(B"
    (its-define-hangul	"wnjT"	"$(CAa(B"	kjhynmlop))

  (its-defrule-hangul	"wnp"	"$(CAb(B")

  (its-defrule-hangul-2	"wnl"	"$(CAc(B"
    (its-define-hangul	"wnlr"	"$(CAd(B"	kijuhynbmlopOP)
    (its-define-hangul	"wnls"	"$(CAe(B"	kijuhynbmlopP)
    (its-define-hangul	"wnlf"	"$(CAf(B"	kijuhynbmlopP)
    (its-define-hangul	"wnla"	"$(CAg(B"	kijuhynbmlopP)
    (its-define-hangul	"wnlq"	"$(CAh(B"	kijuhynbmlopP)
    (its-define-hangul	"wnlt"	"$(CAi(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wb"	"$(CAj(B"
    (its-define-hangul	"wbs"	"$(CAk(B"	kijuhynbmlopP)
    (its-define-hangul	"wbf"	"$(CAl(B"	kijuhynbmlopP)
    (its-define-hangul	"wba"	"$(CAm(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"wm"	"$(CAn(B"
    (its-define-hangul	"wmr"	"$(CAo(B"	kijuhynbmlopOP)
    (its-define-hangul	"wms"	"$(CAp(B"	kijuhynbmlopP)
    (its-define-hangul	"wmf"	"$(CAq(B"	kijuhynbmlopP)
    (its-define-hangul	"wma"	"$(CAr(B"	kijuhynbmlopP)
    (its-define-hangul	"wmq"	"$(CAs(B"	kijuhynbmlopP)
    (its-define-hangul	"wmt"	"$(CAt(B"	kijuhynbmlopOP)
    (its-define-hangul	"wmd"	"$(CAu(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"wl"	"$(CAv(B"
    (its-define-hangul	"wlr"	"$(CAw(B"	kijuhynbmlopOP)
    (its-define-hangul	"wls"	"$(CAx(B"	kijuhynbmlopP)
    (its-define-hangul	"wle"	"$(CAy(B"	kijuhynbmlopP)
    (its-define-hangul	"wlf"	"$(CAz(B"	kijuhynbmlopP
     (its-define-hangul	"wlfa"	"$(CA{(B"	kijuhynbmlopP))
    (its-define-hangul	"wla"	"$(CA|(B"	kijuhynbmlopP)
    (its-define-hangul	"wlq"	"$(CA}(B"	kijuhynbmlopP)
    (its-define-hangul	"wlt"	"$(CA~(B"	kijuhynbmlopOP)
    (its-define-hangul	"wld"	"$(CB!(B"	kijuhynbmlopOP)
    (its-define-hangul	"wlw"	"$(CB"(B"	kijuhynbmlopOP)
    (its-define-hangul	"wlx"	"$(CB#(B"	kijuhynbmlopP)
    (its-define-hangul	"wlv"	"$(CB$(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Wk"	"$(CB%(B"
    (its-define-hangul	"Wkr"	"$(CB&(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wks"	"$(CB'(B"	kijuhynbmlopP
     (its-define-hangul	"Wksg"	"$(CB((B"	kijuhynbmlopP))
    (its-define-hangul	"Wkf"	"$(CB)(B"	kijuhynbmlopP
     (its-define-hangul	"Wkfq"	"$(CB*(B"	kijuhynbmlopP))
    (its-define-hangul	"Wka"	"$(CB+(B"	kijuhynbmlopP)
    (its-define-hangul	"Wkq"	"$(CB,(B"	kijuhynbmlopP)
    (its-define-hangul	"Wkt"	"$(CB-(B"	kijuhynbmlopOP)
    (its-define-hangul	"WkT"	"$(CB.(B"	kjhynmlop)
    (its-define-hangul	"Wkd"	"$(CB/(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wo"	"$(CB0(B"
    (its-define-hangul	"Wor"	"$(CB1(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wos"	"$(CB2(B"	kijuhynbmlopP)
    (its-define-hangul	"Wof"	"$(CB3(B"	kijuhynbmlopP)
    (its-define-hangul	"Woa"	"$(CB4(B"	kijuhynbmlopP)
    (its-define-hangul	"Woq"	"$(CB5(B"	kijuhynbmlopP)
    (its-define-hangul	"Wot"	"$(CB6(B"	kijuhynbmlopOP)
    (its-define-hangul	"WoT"	"$(CB7(B"	kjhynmlop)
    (its-define-hangul	"Wod"	"$(CB8(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wi"	"$(CB9(B"
    (its-define-hangul	"Wis"	"$(CB:(B"	kijuhynbmlopP)
    (its-define-hangul	"Wid"	"$(CB;(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wj"	"$(CB<(B"
    (its-define-hangul	"Wjr"	"$(CB=(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wjs"	"$(CB>(B"	kijuhynbmlopP)
    (its-define-hangul	"Wjf"	"$(CB?(B"	kijuhynbmlopP)
    (its-define-hangul	"Wja"	"$(CB@(B"	kijuhynbmlopP)
    (its-define-hangul	"Wjq"	"$(CBA(B"	kijuhynbmlopP)
    (its-define-hangul	"Wjt"	"$(CBB(B"	kijuhynbmlopOP)
    (its-define-hangul	"WjT"	"$(CBC(B"	kjhynmlop)
    (its-define-hangul	"Wjd"	"$(CBD(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wp"	"$(CBE(B"
    (its-define-hangul	"Wpd"	"$(CBF(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wu"	"$(CBG(B"
    (its-define-hangul	"WuT"	"$(CBH(B"	kjhynmlop))

  (its-defrule-hangul-2	"Wh"	"$(CBI(B"
    (its-define-hangul	"Whr"	"$(CBJ(B"	kijuhynbmlopOP)
    (its-define-hangul	"Whs"	"$(CBK(B"	kijuhynbmlopP)
    (its-define-hangul	"Whf"	"$(CBL(B"	kijuhynbmlopP)
    (its-define-hangul	"Wha"	"$(CBM(B"	kijuhynbmlopP)
    (its-define-hangul	"Whq"	"$(CBN(B"	kijuhynbmlopP)
    (its-define-hangul	"Wht"	"$(CBO(B"	kijuhynbmlopOP)
    (its-define-hangul	"Whd"	"$(CBP(B"	kijuhynbmlopOP)
    (its-define-hangul	"Whc"	"$(CBQ(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"Whk"	"$(CBR(B"
    (its-define-hangul	"Whkr"	"$(CBS(B"	kijuhynbmlopOP)
    (its-define-hangul	"Whkf"	"$(CBT(B"	kijuhynbmlopP)
    (its-define-hangul	"WhkT"	"$(CBU(B"	kjhynmlop))

  (its-defrule-hangul-2	"Who"	"$(CBV(B"
    (its-define-hangul	"WhoT"	"$(CBW(B"	kjhynmlop))

  (its-defrule-hangul-2	"Whl"	"$(CBX(B"
    (its-define-hangul	"Whls"	"$(CBY(B"	kijuhynbmlopP)
    (its-define-hangul	"Whlf"	"$(CBZ(B"	kijuhynbmlopP)
    (its-define-hangul	"Whla"	"$(CB[(B"	kijuhynbmlopP)
    (its-define-hangul	"Whlq"	"$(CB\(B"	kijuhynbmlopP))

  (its-defrule-hangul	"Wy"	"$(C$9(By")
  (its-defrule-hangul	"Wyd"	"$(CB](B")

  (its-defrule-hangul-2	"Wn"	"$(CB^(B"
    (its-define-hangul	"Wnr"	"$(CB_(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wns"	"$(CB`(B"	kijuhynbmlopP)
    (its-define-hangul	"Wnf"	"$(CBa(B"	kijuhynbmlopP)
    (its-define-hangul	"Wna"	"$(CBb(B"	kijuhynbmlopP)
    (its-define-hangul	"Wnq"	"$(CBc(B"	kijuhynbmlopP)
    (its-define-hangul	"Wnd"	"$(CBd(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wnj"	"$(CBe(B"
    (its-define-hangul	"WnjT"	"$(CBf(B"	kjhynmlop)
    (its-define-hangul	"Wnjd"	"$(CBg(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"Wnl"	"$(CBh(B")

  (its-defrule-hangul	"Wb"	"$(CBi(B")

  (its-defrule-hangul-2	"Wm"	"$(CBj(B"
    (its-define-hangul	"Wma"	"$(CBk(B"	kijuhynbmlopP)
    (its-define-hangul	"Wmt"	"$(CBl(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wmd"	"$(CBm(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"Wl"	"$(CBn(B"
    (its-define-hangul	"Wlr"	"$(CBo(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wls"	"$(CBp(B"	kijuhynbmlopP)
    (its-define-hangul	"Wlf"	"$(CBq(B"	kijuhynbmlopP)
    (its-define-hangul	"Wla"	"$(CBr(B"	kijuhynbmlopP)
    (its-define-hangul	"Wlq"	"$(CBs(B"	kijuhynbmlopP)
    (its-define-hangul	"Wld"	"$(CBt(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wlw"	"$(CBu(B"	kijuhynbmlopOP)
    (its-define-hangul	"Wlg"	"$(CBv(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ck"	"$(CBw(B"
    (its-define-hangul	"ckr"	"$(CBx(B"	kijuhynbmlopOP)
    (its-define-hangul	"cks"	"$(CBy(B"	kijuhynbmlopP
     (its-define-hangul	"cksg"	"$(CBz(B"	kijuhynbmlopP))
    (its-define-hangul	"ckf"	"$(CB{(B"	kijuhynbmlopP)
    (its-define-hangul	"cka"	"$(CB|(B"	kijuhynbmlopP)
    (its-define-hangul	"ckq"	"$(CB}(B"	kijuhynbmlopP)
    (its-define-hangul	"ckt"	"$(CB~(B"	kijuhynbmlopOP)
    (its-define-hangul	"ckT"	"$(CC!(B"	kjhynmlop)
    (its-define-hangul	"ckd"	"$(CC"(B"	kijuhynbmlopOP)
    (its-define-hangul	"ckw"	"$(CC#(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"co"	"$(CC$(B"
    (its-define-hangul	"cor"	"$(CC%(B"	kijuhynbmlopOP)
    (its-define-hangul	"cos"	"$(CC&(B"	kijuhynbmlopP)
    (its-define-hangul	"cof"	"$(CC'(B"	kijuhynbmlopP)
    (its-define-hangul	"coa"	"$(CC((B"	kijuhynbmlopP)
    (its-define-hangul	"coq"	"$(CC)(B"	kijuhynbmlopP)
    (its-define-hangul	"cot"	"$(CC*(B"	kijuhynbmlopOP)
    (its-define-hangul	"coT"	"$(CC+(B"	kjhynmlop)
    (its-define-hangul	"cod"	"$(CC,(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ci"	"$(CC-(B"
    (its-define-hangul	"cis"	"$(CC.(B"	kijuhynbmlopP
     (its-define-hangul	"cisg"	"$(CC/(B"	kijuhynbmlopP))
    (its-define-hangul	"cif"	"$(CC0(B"	kijuhynbmlopP)
    (its-define-hangul	"cia"	"$(CC1(B"	kijuhynbmlopP)
    (its-define-hangul	"cid"	"$(CC2(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cj"	"$(CC3(B"
    (its-define-hangul	"cjr"	"$(CC4(B"	kijuhynbmlopOP)
    (its-define-hangul	"cjs"	"$(CC5(B"	kijuhynbmlopP)
    (its-define-hangul	"cjf"	"$(CC6(B"	kijuhynbmlopP)
    (its-define-hangul	"cja"	"$(CC7(B"	kijuhynbmlopP)
    (its-define-hangul	"cjq"	"$(CC8(B"	kijuhynbmlopP)
    (its-define-hangul	"cjt"	"$(CC9(B"	kijuhynbmlopOP)
    (its-define-hangul	"cjT"	"$(CC:(B"	kjhynmlop)
    (its-define-hangul	"cjd"	"$(CC;(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cp"	"$(CC<(B"
    (its-define-hangul	"cpr"	"$(CC=(B"	kijuhynbmlopOP)
    (its-define-hangul	"cps"	"$(CC>(B"	kijuhynbmlopP)
    (its-define-hangul	"cpf"	"$(CC?(B"	kijuhynbmlopP)
    (its-define-hangul	"cpa"	"$(CC@(B"	kijuhynbmlopP)
    (its-define-hangul	"cpq"	"$(CCA(B"	kijuhynbmlopP)
    (its-define-hangul	"cpt"	"$(CCB(B"	kijuhynbmlopOP)
    (its-define-hangul	"cpd"	"$(CCC(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cu"	"$(CCD(B"
    (its-define-hangul	"cus"	"$(CCE(B"	kijuhynbmlopP)
    (its-define-hangul	"cuT"	"$(CCF(B"	kjhynmlop))

  (its-defrule-hangul-2	"cP"	"$(CCG(B"
    (its-define-hangul	"cPs"	"$(CCH(B"	kijuhynbmlopP)
    (its-define-hangul	"cPd"	"$(CCI(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ch"	"$(CCJ(B"
    (its-define-hangul	"chr"	"$(CCK(B"	kijuhynbmlopOP)
    (its-define-hangul	"chs"	"$(CCL(B"	kijuhynbmlopP)
    (its-define-hangul	"chf"	"$(CCM(B"	kijuhynbmlopP)
    (its-define-hangul	"cha"	"$(CCN(B"	kijuhynbmlopP)
    (its-define-hangul	"chq"	"$(CCO(B"	kijuhynbmlopP)
    (its-define-hangul	"cht"	"$(CCP(B"	kijuhynbmlopOP)
    (its-define-hangul	"chd"	"$(CCQ(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"chk"	"$(CCR(B"
    (its-define-hangul	"chks"	"$(CCS(B"	kijuhynbmlopP)
    (its-define-hangul	"chkf"	"$(CCT(B"	kijuhynbmlopP)
    (its-define-hangul	"chkd"	"$(CCU(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"chl"	"$(CCV(B"
    (its-define-hangul	"chls"	"$(CCW(B"	kijuhynbmlopP)
    (its-define-hangul	"chlf"	"$(CCX(B"	kijuhynbmlopP)
    (its-define-hangul	"chla"	"$(CCY(B"	kijuhynbmlopP)
    (its-define-hangul	"chlq"	"$(CCZ(B"	kijuhynbmlopP)
    (its-define-hangul	"chlt"	"$(CC[(B"	kijuhynbmlopOP)
    (its-define-hangul	"chld"	"$(CC\(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cy"	"$(CC](B"
    (its-define-hangul	"cya"	"$(CC^(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"cn"	"$(CC_(B"
    (its-define-hangul	"cnr"	"$(CC`(B"	kijuhynbmlopOP)
    (its-define-hangul	"cns"	"$(CCa(B"	kijuhynbmlopP)
    (its-define-hangul	"cnf"	"$(CCb(B"	kijuhynbmlopP)
    (its-define-hangul	"cna"	"$(CCc(B"	kijuhynbmlopP)
    (its-define-hangul	"cnq"	"$(CCd(B"	kijuhynbmlopP)
    (its-define-hangul	"cnt"	"$(CCe(B"	kijuhynbmlopOP)
    (its-define-hangul	"cnd"	"$(CCf(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cnj"	"$(CCg(B"
    (its-define-hangul	"cnjT"	"$(CCh(B"	kjhynmlop))

  (its-defrule-hangul-2	"cnp"	"$(CCi(B"
    (its-define-hangul	"cnps"	"$(CCj(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"cnl"	"$(CCk(B"
    (its-define-hangul	"cnls"	"$(CCl(B"	kijuhynbmlopP)
    (its-define-hangul	"cnlf"	"$(CCm(B"	kijuhynbmlopP)
    (its-define-hangul	"cnla"	"$(CCn(B"	kijuhynbmlopP)
    (its-define-hangul	"cnlq"	"$(CCo(B"	kijuhynbmlopP)
    (its-define-hangul	"cnlt"	"$(CCp(B"	kijuhynbmlopOP)
    (its-define-hangul	"cnld"	"$(CCq(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cb"	"$(CCr(B"
    (its-define-hangul	"cbs"	"$(CCs(B"	kijuhynbmlopP)
    (its-define-hangul	"cbf"	"$(CCt(B"	kijuhynbmlopP)
    (its-define-hangul	"cba"	"$(CCu(B"	kijuhynbmlopP)
    (its-define-hangul	"cbd"	"$(CCv(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cm"	"$(CCw(B"
    (its-define-hangul	"cmr"	"$(CCx(B"	kijuhynbmlopOP)
    (its-define-hangul	"cms"	"$(CCy(B"	kijuhynbmlopP)
    (its-define-hangul	"cmf"	"$(CCz(B"	kijuhynbmlopP)
    (its-define-hangul	"cma"	"$(CC{(B"	kijuhynbmlopP)
    (its-define-hangul	"cmq"	"$(CC|(B"	kijuhynbmlopP)
    (its-define-hangul	"cmt"	"$(CC}(B"	kijuhynbmlopOP)
    (its-define-hangul	"cmd"	"$(CC~(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"cl"	"$(CD!(B"
    (its-define-hangul	"clr"	"$(CD"(B"	kijuhynbmlopOP)
    (its-define-hangul	"cls"	"$(CD#(B"	kijuhynbmlopP)
    (its-define-hangul	"cle"	"$(CD$(B"	kijuhynbmlopP)
    (its-define-hangul	"clf"	"$(CD%(B"	kijuhynbmlopP
     (its-define-hangul	"clfr"	"$(CD&(B"	kijuhynbmlopOP))
    (its-define-hangul	"cla"	"$(CD'(B"	kijuhynbmlopP)
    (its-define-hangul	"clq"	"$(CD((B"	kijuhynbmlopP)
    (its-define-hangul	"clt"	"$(CD)(B"	kijuhynbmlopOP)
    (its-define-hangul	"cld"	"$(CD*(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zk"	"$(CD+(B"
    (its-define-hangul	"zkr"	"$(CD,(B"	kijuhynbmlopOP)
    (its-define-hangul	"zks"	"$(CD-(B"	kijuhynbmlopP)
    (its-define-hangul	"zkf"	"$(CD.(B"	kijuhynbmlopP)
    (its-define-hangul	"zka"	"$(CD/(B"	kijuhynbmlopP)
    (its-define-hangul	"zkq"	"$(CD0(B"	kijuhynbmlopP)
    (its-define-hangul	"zkt"	"$(CD1(B"	kijuhynbmlopOP)
    (its-define-hangul	"zkd"	"$(CD2(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zo"	"$(CD3(B"
    (its-define-hangul	"zor"	"$(CD4(B"	kijuhynbmlopOP)
    (its-define-hangul	"zos"	"$(CD5(B"	kijuhynbmlopP)
    (its-define-hangul	"zof"	"$(CD6(B"	kijuhynbmlopP)
    (its-define-hangul	"zoa"	"$(CD7(B"	kijuhynbmlopP)
    (its-define-hangul	"zoq"	"$(CD8(B"	kijuhynbmlopP)
    (its-define-hangul	"zot"	"$(CD9(B"	kijuhynbmlopOP)
    (its-define-hangul	"zoT"	"$(CD:(B"	kjhynmlop)
    (its-define-hangul	"zod"	"$(CD;(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zi"	"$(CD<(B"
    (its-define-hangul	"zir"	"$(CD=(B"	kijuhynbmlopOP)
    (its-define-hangul	"zid"	"$(CD>(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zj"	"$(CD?(B"
    (its-define-hangul	"zjr"	"$(CD@(B"	kijuhynbmlopOP)
    (its-define-hangul	"zjs"	"$(CDA(B"	kijuhynbmlopP)
    (its-define-hangul	"zje"	"$(CDB(B"	kijuhynbmlopP)
    (its-define-hangul	"zjf"	"$(CDC(B"	kijuhynbmlopP)
    (its-define-hangul	"zja"	"$(CDD(B"	kijuhynbmlopP)
    (its-define-hangul	"zjq"	"$(CDE(B"	kijuhynbmlopP)
    (its-define-hangul	"zjt"	"$(CDF(B"	kijuhynbmlopOP)
    (its-define-hangul	"zjT"	"$(CDG(B"	kjhynmlop)
    (its-define-hangul	"zjd"	"$(CDH(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zp"	"$(CDI(B"
    (its-define-hangul	"zpr"	"$(CDJ(B"	kijuhynbmlopOP)
    (its-define-hangul	"zps"	"$(CDK(B"	kijuhynbmlopP)
    (its-define-hangul	"zpf"	"$(CDL(B"	kijuhynbmlopP)
    (its-define-hangul	"zpa"	"$(CDM(B"	kijuhynbmlopP)
    (its-define-hangul	"zpq"	"$(CDN(B"	kijuhynbmlopP)
    (its-define-hangul	"zpt"	"$(CDO(B"	kijuhynbmlopOP)
    (its-define-hangul	"zpd"	"$(CDP(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zu"	"$(CDQ(B"
    (its-define-hangul	"zus"	"$(CDR(B"	kijuhynbmlopP)
    (its-define-hangul	"zuf"	"$(CDS(B"	kijuhynbmlopP)
    (its-define-hangul	"zua"	"$(CDT(B"	kijuhynbmlopP)
    (its-define-hangul	"zuq"	"$(CDU(B"	kijuhynbmlopP)
    (its-define-hangul	"zut"	"$(CDV(B"	kijuhynbmlopOP)
    (its-define-hangul	"zuT"	"$(CDW(B"	kjhynmlop)
    (its-define-hangul	"zud"	"$(CDX(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"zP"	"$(CDY(B")

  (its-defrule-hangul-2	"zh"	"$(CDZ(B"
    (its-define-hangul	"zhr"	"$(CD[(B"	kijuhynbmlopOP)
    (its-define-hangul	"zhs"	"$(CD\(B"	kijuhynbmlopP)
    (its-define-hangul	"zhf"	"$(CD](B"	kijuhynbmlopP)
    (its-define-hangul	"zha"	"$(CD^(B"	kijuhynbmlopP)
    (its-define-hangul	"zhq"	"$(CD_(B"	kijuhynbmlopP)
    (its-define-hangul	"zht"	"$(CD`(B"	kijuhynbmlopOP)
    (its-define-hangul	"zhd"	"$(CDa(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zhk"	"$(CDb(B"
    (its-define-hangul	"zhkr"	"$(CDc(B"	kijuhynbmlopOP)
    (its-define-hangul	"zhks"	"$(CDd(B"	kijuhynbmlopP)
    (its-define-hangul	"zhkf"	"$(CDe(B"	kijuhynbmlopP)
    (its-define-hangul	"zhka"	"$(CDf(B"	kijuhynbmlopP)
    (its-define-hangul	"zhkd"	"$(CDg(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zho"	"$(CDh(B"
    (its-define-hangul	"zhod"	"$(CDi(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zhl"	"$(CDj(B"
    (its-define-hangul	"zhlf"	"$(CDk(B"	kijuhynbmlopP))

  (its-defrule-hangul	"zy"	"$(CDl(B")

  (its-defrule-hangul-2	"zn"	"$(CDm(B"
    (its-define-hangul	"znr"	"$(CDn(B"	kijuhynbmlopOP)
    (its-define-hangul	"znf"	"$(CDp(B"	kijuhynbmlopP)
    (its-define-hangul	"zna"	"$(CDq(B"	kijuhynbmlopP)
    (its-define-hangul	"znq"	"$(CDr(B"	kijuhynbmlopP)
    (its-define-hangul	"znt"	"$(CDs(B"	kijuhynbmlopOP)
    (its-define-hangul	"znd"	"$(CDt(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"znj"	"$(CDu(B"
    (its-define-hangul	"znjs"	"$(CDv(B"	kijuhynbmlopP)
    (its-define-hangul	"znjf"	"$(CDw(B"	kijuhynbmlopP)
    (its-define-hangul	"znjd"	"$(CDx(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"znp"	"$(CDy(B"
    (its-define-hangul	"znpd"	"$(CDz(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"znl"	"$(CD{(B"
    (its-define-hangul	"znlr"	"$(CD|(B"	kijuhynbmlopOP)
    (its-define-hangul	"znls"	"$(CD}(B"	kijuhynbmlopP)
    (its-define-hangul	"znlf"	"$(CD~(B"	kijuhynbmlopP)
    (its-define-hangul	"znla"	"$(CE!(B"	kijuhynbmlopP)
    (its-define-hangul	"znlq"	"$(CE"(B"	kijuhynbmlopP)
    (its-define-hangul	"znlt"	"$(CE#(B"	kijuhynbmlopOP)
    (its-define-hangul	"znld"	"$(CE$(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zb"	"$(CE%(B"
    (its-define-hangul	"zbs"	"$(CE&(B"	kijuhynbmlopP)
    (its-define-hangul	"zbf"	"$(CE'(B"	kijuhynbmlopP)
    (its-define-hangul	"zba"	"$(CE((B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"zm"	"$(CE)(B"
    (its-define-hangul	"zmr"	"$(CE*(B"	kijuhynbmlopOP)
    (its-define-hangul	"zms"	"$(CE+(B"	kijuhynbmlopP)
    (its-define-hangul	"zmf"	"$(CE,(B"	kijuhynbmlopP)
    (its-define-hangul	"zma"	"$(CE-(B"	kijuhynbmlopP)
    (its-define-hangul	"zmq"	"$(CE.(B"	kijuhynbmlopP)
    (its-define-hangul	"zmd"	"$(CE/(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"zl"	"$(CE0(B"
    (its-define-hangul	"zlr"	"$(CE1(B"	kijuhynbmlopOP)
    (its-define-hangul	"zls"	"$(CE2(B"	kijuhynbmlopP)
    (its-define-hangul	"zlf"	"$(CE3(B"	kijuhynbmlopP)
    (its-define-hangul	"zla"	"$(CE4(B"	kijuhynbmlopP)
    (its-define-hangul	"zlq"	"$(CE5(B"	kijuhynbmlopP)
    (its-define-hangul	"zlt"	"$(CE6(B"	kijuhynbmlopOP)
    (its-define-hangul	"zld"	"$(CE7(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xk"	"$(CE8(B"
    (its-define-hangul	"xkr"	"$(CE9(B"	kijuhynbmlopOP)
    (its-define-hangul	"xks"	"$(CE:(B"	kijuhynbmlopP)
    (its-define-hangul	"xkf"	"$(CE;(B"	kijuhynbmlopP
     (its-define-hangul	"xkfr"	"$(CE<(B"	kijuhynbmlopOP))
    (its-define-hangul	"xka"	"$(CE=(B"	kijuhynbmlopP)
    (its-define-hangul	"xkq"	"$(CE>(B"	kijuhynbmlopP)
    (its-define-hangul	"xkt"	"$(CE?(B"	kijuhynbmlopOP)
    (its-define-hangul	"xkT"	"$(CE@(B"	kjhynmlop)
    (its-define-hangul	"xkd"	"$(CEA(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xo"	"$(CEB(B"
    (its-define-hangul	"xor"	"$(CEC(B"	kijuhynbmlopOP)
    (its-define-hangul	"xos"	"$(CED(B"	kijuhynbmlopP)
    (its-define-hangul	"xof"	"$(CEE(B"	kijuhynbmlopP)
    (its-define-hangul	"xoa"	"$(CEF(B"	kijuhynbmlopP)
    (its-define-hangul	"xoq"	"$(CEG(B"	kijuhynbmlopP)
    (its-define-hangul	"xot"	"$(CEH(B"	kijuhynbmlopOP)
    (its-define-hangul	"xoT"	"$(CEI(B"	kjhynmlop)
    (its-define-hangul	"xod"	"$(CEJ(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xi"	"$(CEK(B"
    (its-define-hangul	"xid"	"$(CEL(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xj"	"$(CEM(B"
    (its-define-hangul	"xjr"	"$(CEN(B"	kijuhynbmlopOP)
    (its-define-hangul	"xjs"	"$(CEO(B"	kijuhynbmlopP)
    (its-define-hangul	"xjf"	"$(CEP(B"	kijuhynbmlopP
     (its-define-hangul	"xjfa"	"$(CEQ(B"	kijuhynbmlopP))
    (its-define-hangul	"xja"	"$(CER(B"	kijuhynbmlopP)
    (its-define-hangul	"xjq"	"$(CES(B"	kijuhynbmlopP)
    (its-define-hangul	"xjt"	"$(CET(B"	kijuhynbmlopOP)
    (its-define-hangul	"xjT"	"$(CEU(B"	kjhynmlop)
    (its-define-hangul	"xjd"	"$(CEV(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xp"	"$(CEW(B"
    (its-define-hangul	"xpr"	"$(CEX(B"	kijuhynbmlopOP)
    (its-define-hangul	"xps"	"$(CEY(B"	kijuhynbmlopP)
    (its-define-hangul	"xpf"	"$(CEZ(B"	kijuhynbmlopP)
    (its-define-hangul	"xpa"	"$(CE[(B"	kijuhynbmlopP)
    (its-define-hangul	"xpq"	"$(CE\(B"	kijuhynbmlopP)
    (its-define-hangul	"xpt"	"$(CE](B"	kijuhynbmlopOP)
    (its-define-hangul	"xpd"	"$(CE^(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xu"	"$(CE_(B"
    (its-define-hangul	"xus"	"$(CE`(B"	kijuhynbmlopP)
    (its-define-hangul	"xuT"	"$(CEa(B"	kjhynmlop))

  (its-defrule-hangul-2	"xP"	"$(CEb(B"
    (its-define-hangul	"xPs"	"$(CEc(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"xh"	"$(CEd(B"
    (its-define-hangul	"xhr"	"$(CEe(B"	kijuhynbmlopOP)
    (its-define-hangul	"xhs"	"$(CEf(B"	kijuhynbmlopP)
    (its-define-hangul	"xhf"	"$(CEg(B"	kijuhynbmlopP)
    (its-define-hangul	"xha"	"$(CEh(B"	kijuhynbmlopP)
    (its-define-hangul	"xhq"	"$(CEi(B"	kijuhynbmlopP)
    (its-define-hangul	"xht"	"$(CEj(B"	kijuhynbmlopOP)
    (its-define-hangul	"xhd"	"$(CEk(B"	kijuhynbmlopOP)
    (its-define-hangul	"xhv"	"$(CEl(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"xhk"	"$(CEm(B"
    (its-define-hangul	"xhks"	"$(CEn(B"	kijuhynbmlopP))

  (its-defrule-hangul	"xho"	"$(CEo(B")

  (its-defrule-hangul-2	"xhl"	"$(CEp(B"
    (its-define-hangul	"xhls"	"$(CEq(B"	kijuhynbmlopP)
    (its-define-hangul	"xhlt"	"$(CEr(B"	kijuhynbmlopOP)
    (its-define-hangul	"xhld"	"$(CEs(B"	kijuhynbmlopOP))

  (its-defrule-hangul	"xy"	"$(CEt(B")

  (its-defrule-hangul-2	"xn"	"$(CEu(B"
    (its-define-hangul	"xnr"	"$(CEv(B"	kijuhynbmlopOP)
    (its-define-hangul	"xns"	"$(CEw(B"	kijuhynbmlopP)
    (its-define-hangul	"xnf"	"$(CEx(B"	kijuhynbmlopP)
    (its-define-hangul	"xna"	"$(CEy(B"	kijuhynbmlopP)
    (its-define-hangul	"xnq"	"$(CEz(B"	kijuhynbmlopP)
    (its-define-hangul	"xnt"	"$(CE{(B"	kijuhynbmlopOP)
    (its-define-hangul	"xnd"	"$(CE|(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xnj"	"$(CE}(B"
    (its-define-hangul	"xnjT"	"$(CE~(B"	kjhynmlop))

  (its-defrule-hangul	"xnp"	"$(CF!(B")

  (its-defrule-hangul-2	"xnl"	"$(CF"(B"
    (its-define-hangul	"xnlr"	"$(CF#(B"	kijuhynbmlopOP)
    (its-define-hangul	"xnls"	"$(CF$(B"	kijuhynbmlopP)
    (its-define-hangul	"xnlf"	"$(CF%(B"	kijuhynbmlopP)
    (its-define-hangul	"xnla"	"$(CF&(B"	kijuhynbmlopP)
    (its-define-hangul	"xnlq"	"$(CF'(B"	kijuhynbmlopP)
    (its-define-hangul	"xnld"	"$(CF((B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xb"	"$(CF)(B"
    (its-define-hangul	"xbs"	"$(CF*(B"	kijuhynbmlopP)
    (its-define-hangul	"xbf"	"$(CF+(B"	kijuhynbmlopP)
    (its-define-hangul	"xba"	"$(CF,(B"	kijuhynbmlopP)
    (its-define-hangul	"xbd"	"$(CF-(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xm"	"$(CF.(B"
    (its-define-hangul	"xmr"	"$(CF/(B"	kijuhynbmlopOP)
    (its-define-hangul	"xms"	"$(CF0(B"	kijuhynbmlopP)
    (its-define-hangul	"xme"	"$(CF1(B"	kijuhynbmlopP)
    (its-define-hangul	"xmf"	"$(CF2(B"	kijuhynbmlopP
     (its-define-hangul	"xmfa"	"$(CF3(B"	kijuhynbmlopP))
    (its-define-hangul	"xma"	"$(CF4(B"	kijuhynbmlopP)
    (its-define-hangul	"xmq"	"$(CF5(B"	kijuhynbmlopP)
    (its-define-hangul	"xmt"	"$(CF6(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"xml"	"$(CF7(B"
    (its-define-hangul	"xmls"	"$(CF8(B"	kijuhynbmlopP)
    (its-define-hangul	"xmlf"	"$(CF9(B"	kijuhynbmlopP)
    (its-define-hangul	"xmla"	"$(CF:(B"	kijuhynbmlopP)
    (its-define-hangul	"xmlq"	"$(CF;(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"xl"	"$(CF<(B"
    (its-define-hangul	"xlr"	"$(CF=(B"	kijuhynbmlopOP)
    (its-define-hangul	"xls"	"$(CF>(B"	kijuhynbmlopP)
    (its-define-hangul	"xlf"	"$(CF?(B"	kijuhynbmlopP)
    (its-define-hangul	"xla"	"$(CF@(B"	kijuhynbmlopP)
    (its-define-hangul	"xlq"	"$(CFA(B"	kijuhynbmlopP)
    (its-define-hangul	"xlt"	"$(CFB(B"	kijuhynbmlopOP)
    (its-define-hangul	"xld"	"$(CFC(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vk"	"$(CFD(B"
    (its-define-hangul	"vkr"	"$(CFE(B"	kijuhynbmlopOP)
    (its-define-hangul	"vkR"	"$(CFF(B"	kijuhynbmlopP)
    (its-define-hangul	"vks"	"$(CFG(B"	kijuhynbmlopP)
    (its-define-hangul	"vkf"	"$(CFH(B"	kijuhynbmlopP
     (its-define-hangul	"vkfa"	"$(CFI(B"	kijuhynbmlopP))
    (its-define-hangul	"vka"	"$(CFJ(B"	kijuhynbmlopP)
    (its-define-hangul	"vkq"	"$(CFK(B"	kijuhynbmlopP)
    (its-define-hangul	"vkt"	"$(CFL(B"	kijuhynbmlopOP)
    (its-define-hangul	"vkT"	"$(CFM(B"	kjhynmlop)
    (its-define-hangul	"vkd"	"$(CFN(B"	kijuhynbmlopOP)
    (its-define-hangul	"vkx"	"$(CFO(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"vo"	"$(CFP(B"
    (its-define-hangul	"vor"	"$(CFQ(B"	kijuhynbmlopOP)
    (its-define-hangul	"vos"	"$(CFR(B"	kijuhynbmlopP)
    (its-define-hangul	"vof"	"$(CFS(B"	kijuhynbmlopP)
    (its-define-hangul	"voa"	"$(CFT(B"	kijuhynbmlopP)
    (its-define-hangul	"voq"	"$(CFU(B"	kijuhynbmlopP)
    (its-define-hangul	"vot"	"$(CFV(B"	kijuhynbmlopOP)
    (its-define-hangul	"voT"	"$(CFW(B"	kjhynmlop)
    (its-define-hangul	"vod"	"$(CFX(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vi"	"$(CFY(B"
    (its-define-hangul	"vir"	"$(CFZ(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vj"	"$(CF[(B"
    (its-define-hangul	"vjr"	"$(CF\(B"	kijuhynbmlopOP)
    (its-define-hangul	"vjs"	"$(CF](B"	kijuhynbmlopP)
    (its-define-hangul	"vja"	"$(CF_(B"	kijuhynbmlopP)
    (its-define-hangul	"vjq"	"$(CF`(B"	kijuhynbmlopP)
    (its-define-hangul	"vjt"	"$(CFa(B"	kijuhynbmlopOP)
    (its-define-hangul	"vjT"	"$(CFb(B"	kjhynmlop)
    (its-define-hangul	"vjd"	"$(CFc(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vp"	"$(CFd(B"
    (its-define-hangul	"vpr"	"$(CFe(B"	kijuhynbmlopOP)
    (its-define-hangul	"vps"	"$(CFf(B"	kijuhynbmlopP)
    (its-define-hangul	"vpf"	"$(CFg(B"	kijuhynbmlopP)
    (its-define-hangul	"vpa"	"$(CFh(B"	kijuhynbmlopP)
    (its-define-hangul	"vpq"	"$(CFi(B"	kijuhynbmlopP)
    (its-define-hangul	"vpt"	"$(CFj(B"	kijuhynbmlopOP)
    (its-define-hangul	"vpd"	"$(CFk(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vu"	"$(CFl(B"
    (its-define-hangul	"vus"	"$(CFm(B"	kijuhynbmlopP)
    (its-define-hangul	"vuf"	"$(CFn(B"	kijuhynbmlopP)
    (its-define-hangul	"vua"	"$(CFo(B"	kijuhynbmlopP)
    (its-define-hangul	"vuq"	"$(CFp(B"	kijuhynbmlopP)
    (its-define-hangul	"vuT"	"$(CFq(B"	kjhynmlop)
    (its-define-hangul	"vud"	"$(CFr(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vP"	"$(CFs(B"
    (its-define-hangul	"vPf"	"$(CFt(B"	kijuhynbmlopP)
    (its-define-hangul	"vPq"	"$(CFu(B"	kijuhynbmlopP)
    (its-define-hangul	"vPt"	"$(CFv(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vh"	"$(CFw(B"
    (its-define-hangul	"vhr"	"$(CFx(B"	kijuhynbmlopOP)
    (its-define-hangul	"vhs"	"$(CFy(B"	kijuhynbmlopP)
    (its-define-hangul	"vhf"	"$(CFz(B"	kijuhynbmlopP)
    (its-define-hangul	"vha"	"$(CF{(B"	kijuhynbmlopP)
    (its-define-hangul	"vhq"	"$(CF|(B"	kijuhynbmlopP)
    (its-define-hangul	"vht"	"$(CF}(B"	kijuhynbmlopOP)
    (its-define-hangul	"vhd"	"$(CF~(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vhk"	"$(CG!(B"
    (its-define-hangul	"vhkd"	"$(CG"(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vhl"	"$(CG#(B"
    (its-define-hangul	"vhls"	"$(CG$(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"vy"	"$(CG%(B"
    (its-define-hangul	"vys"	"$(CG&(B"	kijuhynbmlopP)
    (its-define-hangul	"vyf"	"$(CG'(B"	kijuhynbmlopP)
    (its-define-hangul	"vyq"	"$(CG((B"	kijuhynbmlopP)
    (its-define-hangul	"vyt"	"$(CG)(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vn"	"$(CG*(B"
    (its-define-hangul	"vnr"	"$(CG+(B"	kijuhynbmlopOP)
    (its-define-hangul	"vns"	"$(CG,(B"	kijuhynbmlopP)
    (its-define-hangul	"vne"	"$(CG-(B"	kijuhynbmlopP)
    (its-define-hangul	"vnf"	"$(CG.(B"	kijuhynbmlopP
     (its-define-hangul	"vnfa"	"$(CG/(B"	kijuhynbmlopP))
    (its-define-hangul	"vna"	"$(CG0(B"	kijuhynbmlopP)
    (its-define-hangul	"vnq"	"$(CG1(B"	kijuhynbmlopP)
    (its-define-hangul	"vnt"	"$(CG2(B"	kijuhynbmlopOP)
    (its-define-hangul	"vnd"	"$(CG3(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vnj"	"$(CG4(B"
    (its-define-hangul	"vnjd"	"$(CG5(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vnl"	"$(CG6(B"
    (its-define-hangul	"vnls"	"$(CG7(B"	kijuhynbmlopP)
    (its-define-hangul	"vnlf"	"$(CG8(B"	kijuhynbmlopP)
    (its-define-hangul	"vnla"	"$(CG9(B"	kijuhynbmlopP)
    (its-define-hangul	"vnlt"	"$(CG:(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vb"	"$(CG;(B"
    (its-define-hangul	"vbs"	"$(CG<(B"	kijuhynbmlopP)
    (its-define-hangul	"vbf"	"$(CG=(B"	kijuhynbmlopP)
    (its-define-hangul	"vba"	"$(CG>(B"	kijuhynbmlopP)
    (its-define-hangul	"vbt"	"$(CG?(B"	kijuhynbmlopOP)
    (its-define-hangul	"vbd"	"$(CG@(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vm"	"$(CGA(B"
    (its-define-hangul	"vms"	"$(CGB(B"	kijuhynbmlopP)
    (its-define-hangul	"vmf"	"$(CGC(B"	kijuhynbmlopP)
    (its-define-hangul	"vma"	"$(CGD(B"	kijuhynbmlopP)
    (its-define-hangul	"vmq"	"$(CGE(B"	kijuhynbmlopP)
    (its-define-hangul	"vmt"	"$(CGF(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"vl"	"$(CGG(B"
    (its-define-hangul	"vlr"	"$(CGH(B"	kijuhynbmlopOP)
    (its-define-hangul	"vls"	"$(CGI(B"	kijuhynbmlopP)
    (its-define-hangul	"vlf"	"$(CGJ(B"	kijuhynbmlopP)
    (its-define-hangul	"vla"	"$(CGK(B"	kijuhynbmlopP)
    (its-define-hangul	"vlq"	"$(CGL(B"	kijuhynbmlopP)
    (its-define-hangul	"vlt"	"$(CGM(B"	kijuhynbmlopOP)
    (its-define-hangul	"vld"	"$(CGN(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gk"	"$(CGO(B"
    (its-define-hangul	"gkr"	"$(CGP(B"	kijuhynbmlopOP)
    (its-define-hangul	"gks"	"$(CGQ(B"	kijuhynbmlopP)
    (its-define-hangul	"gkf"	"$(CGR(B"	kijuhynbmlopP
     (its-define-hangul	"gkfx"	"$(CGS(B"	kijuhynbmlopP))
    (its-define-hangul	"gka"	"$(CGT(B"	kijuhynbmlopP)
    (its-define-hangul	"gkq"	"$(CGU(B"	kijuhynbmlopP)
    (its-define-hangul	"gkt"	"$(CGV(B"	kijuhynbmlopOP)
    (its-define-hangul	"gkd"	"$(CGW(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"go"	"$(CGX(B"
    (its-define-hangul	"gor"	"$(CGY(B"	kijuhynbmlopOP)
    (its-define-hangul	"gos"	"$(CGZ(B"	kijuhynbmlopP)
    (its-define-hangul	"gof"	"$(CG[(B"	kijuhynbmlopP)
    (its-define-hangul	"goa"	"$(CG\(B"	kijuhynbmlopP)
    (its-define-hangul	"goq"	"$(CG](B"	kijuhynbmlopP)
    (its-define-hangul	"got"	"$(CG^(B"	kijuhynbmlopOP)
    (its-define-hangul	"goT"	"$(CG_(B"	kjhynmlop)
    (its-define-hangul	"god"	"$(CG`(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gi"	"$(CGa(B"
    (its-define-hangul	"gid"	"$(CGb(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gj"	"$(CGc(B"
    (its-define-hangul	"gjr"	"$(CGd(B"	kijuhynbmlopOP)
    (its-define-hangul	"gjs"	"$(CGe(B"	kijuhynbmlopP)
    (its-define-hangul	"gjf"	"$(CGf(B"	kijuhynbmlopP
     (its-define-hangul	"gjfa"	"$(CGg(B"	kijuhynbmlopP))
    (its-define-hangul	"gja"	"$(CGh(B"	kijuhynbmlopP)
    (its-define-hangul	"gjq"	"$(CGi(B"	kijuhynbmlopP)
    (its-define-hangul	"gjt"	"$(CGj(B"	kijuhynbmlopOP)
    (its-define-hangul	"gjd"	"$(CGk(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gp"	"$(CGl(B"
    (its-define-hangul	"gpr"	"$(CGm(B"	kijuhynbmlopOP)
    (its-define-hangul	"gps"	"$(CGn(B"	kijuhynbmlopP)
    (its-define-hangul	"gpf"	"$(CGo(B"	kijuhynbmlopP)
    (its-define-hangul	"gpa"	"$(CGp(B"	kijuhynbmlopP)
    (its-define-hangul	"gpq"	"$(CGq(B"	kijuhynbmlopP)
    (its-define-hangul	"gpt"	"$(CGr(B"	kijuhynbmlopOP)
    (its-define-hangul	"gpd"	"$(CGs(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gu"	"$(CGt(B"
    (its-define-hangul	"gur"	"$(CGu(B"	kijuhynbmlopOP)
    (its-define-hangul	"gus"	"$(CGv(B"	kijuhynbmlopP)
    (its-define-hangul	"guf"	"$(CGw(B"	kijuhynbmlopP)
    (its-define-hangul	"gua"	"$(CGx(B"	kijuhynbmlopP)
    (its-define-hangul	"guq"	"$(CGy(B"	kijuhynbmlopP)
    (its-define-hangul	"gut"	"$(CGz(B"	kijuhynbmlopOP)
    (its-define-hangul	"guT"	"$(CG{(B"	kjhynmlop)
    (its-define-hangul	"gud"	"$(CG|(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gP"	"$(CG}(B"
    (its-define-hangul	"gPs"	"$(CG~(B"	kijuhynbmlopP)
    (its-define-hangul	"gPf"	"$(CH!(B"	kijuhynbmlopP)
    (its-define-hangul	"gPq"	"$(CH"(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"gh"	"$(CH#(B"
    (its-define-hangul	"ghr"	"$(CH$(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghs"	"$(CH%(B"	kijuhynbmlopP)
    (its-define-hangul	"ghf"	"$(CH&(B"	kijuhynbmlopP
     (its-define-hangul	"ghfx"	"$(CH'(B"	kijuhynbmlopP))
    (its-define-hangul	"gha"	"$(CH((B"	kijuhynbmlopP)
    (its-define-hangul	"ghq"	"$(CH)(B"	kijuhynbmlopP)
    (its-define-hangul	"ght"	"$(CH*(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghd"	"$(CH+(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghx"	"$(CH,(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"ghk"	"$(CH-(B"
    (its-define-hangul	"ghkr"	"$(CH.(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghks"	"$(CH/(B"	kijuhynbmlopP)
    (its-define-hangul	"ghkf"	"$(CH0(B"	kijuhynbmlopP)
    (its-define-hangul	"ghkt"	"$(CH1(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghkd"	"$(CH2(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gho"	"$(CH3(B"
    (its-define-hangul	"ghor"	"$(CH4(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghos"	"$(CH5(B"	kijuhynbmlopP)
    (its-define-hangul	"ghot"	"$(CH6(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghod"	"$(CH7(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"ghl"	"$(CH8(B"
    (its-define-hangul	"ghlr"	"$(CH9(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghls"	"$(CH:(B"	kijuhynbmlopP)
    (its-define-hangul	"ghlf"	"$(CH;(B"	kijuhynbmlopP)
    (its-define-hangul	"ghlq"	"$(CH<(B"	kijuhynbmlopP)
    (its-define-hangul	"ghlt"	"$(CH=(B"	kijuhynbmlopOP)
    (its-define-hangul	"ghld"	"$(CH>(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gy"	"$(CH?(B"
    (its-define-hangul	"gys"	"$(CH@(B"	kijuhynbmlopP)
    (its-define-hangul	"gyf"	"$(CHA(B"	kijuhynbmlopP)
    (its-define-hangul	"gyq"	"$(CHB(B"	kijuhynbmlopP)
    (its-define-hangul	"gyt"	"$(CHC(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gn"	"$(CHD(B"
    (its-define-hangul	"gnr"	"$(CHE(B"	kijuhynbmlopOP)
    (its-define-hangul	"gns"	"$(CHF(B"	kijuhynbmlopP)
    (its-define-hangul	"gnf"	"$(CHG(B"	kijuhynbmlopP
     (its-define-hangul	"gnfx"	"$(CHH(B"	kijuhynbmlopP))
    (its-define-hangul	"gna"	"$(CHI(B"	kijuhynbmlopP)
    (its-define-hangul	"gnt"	"$(CHJ(B"	kijuhynbmlopOP)
    (its-define-hangul	"gnd"	"$(CHK(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gnj"	"$(CHL(B"
    (its-define-hangul	"gnjs"	"$(CHM(B"	kijuhynbmlopP)
    (its-define-hangul	"gnjf"	"$(CHN(B"	kijuhynbmlopP)
    (its-define-hangul	"gnja"	"$(CHO(B"	kijuhynbmlopP)
    (its-define-hangul	"gnjd"	"$(CHP(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gnp"	"$(CHQ(B"
    (its-define-hangul	"gnpr"	"$(CHR(B"	kijuhynbmlopOP)
    (its-define-hangul	"gnps"	"$(CHS(B"	kijuhynbmlopP)
    (its-define-hangul	"gnpf"	"$(CHT(B"	kijuhynbmlopP)
    (its-define-hangul	"gnpd"	"$(CHU(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gnl"	"$(CHV(B"
    (its-define-hangul	"gnlr"	"$(CHW(B"	kijuhynbmlopOP)
    (its-define-hangul	"gnls"	"$(CHX(B"	kijuhynbmlopP)
    (its-define-hangul	"gnlf"	"$(CHY(B"	kijuhynbmlopP)
    (its-define-hangul	"gnla"	"$(CHZ(B"	kijuhynbmlopP)
    (its-define-hangul	"gnlq"	"$(CH[(B"	kijuhynbmlopP)
    (its-define-hangul	"gnlt"	"$(CH\(B"	kijuhynbmlopOP)
    (its-define-hangul	"gnld"	"$(CH](B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gb"	"$(CH^(B"
    (its-define-hangul	"gbr"	"$(CH_(B"	kijuhynbmlopOP)
    (its-define-hangul	"gbs"	"$(CH`(B"	kijuhynbmlopP)
    (its-define-hangul	"gbf"	"$(CHa(B"	kijuhynbmlopP)
    (its-define-hangul	"gba"	"$(CHb(B"	kijuhynbmlopP)
    (its-define-hangul	"gbt"	"$(CHc(B"	kijuhynbmlopOP)
    (its-define-hangul	"gbd"	"$(CHd(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gm"	"$(CHe(B"
    (its-define-hangul	"gmr"	"$(CHf(B"	kijuhynbmlopOP)
    (its-define-hangul	"gms"	"$(CHg(B"	kijuhynbmlopP
     (its-define-hangul	"gmsg"	"$(CHh(B"	kijuhynbmlopP))
    (its-define-hangul	"gme"	"$(CHi(B"	kijuhynbmlopP)
    (its-define-hangul	"gmf"	"$(CHj(B"	kijuhynbmlopP
     (its-define-hangul	"gmfr"	"$(CHk(B"	kijuhynbmlopOP))
    (its-define-hangul	"gma"	"$(CHl(B"	kijuhynbmlopP)
    (its-define-hangul	"gmq"	"$(CHm(B"	kijuhynbmlopP)
    (its-define-hangul	"gmt"	"$(CHn(B"	kijuhynbmlopOP)
    (its-define-hangul	"gmd"	"$(CHo(B"	kijuhynbmlopOP)
    (its-define-hangul	"gmx"	"$(CHp(B"	kijuhynbmlopP))

  (its-defrule-hangul-2	"gml"	"$(CHq(B"
    (its-define-hangul	"gmls"	"$(CHr(B"	kijuhynbmlopP)
    (its-define-hangul	"gmlf"	"$(CHs(B"	kijuhynbmlopP)
    (its-define-hangul	"gmla"	"$(CHt(B"	kijuhynbmlopP)
    (its-define-hangul	"gmlq"	"$(CHu(B"	kijuhynbmlopP)
    (its-define-hangul	"gmld"	"$(CHv(B"	kijuhynbmlopOP))

  (its-defrule-hangul-2	"gl"	"$(CHw(B"
    (its-define-hangul	"glr"	"$(CHx(B"	kijuhynbmlopOP)
    (its-define-hangul	"gls"	"$(CHy(B"	kijuhynbmlopP)
    (its-define-hangul	"glf"	"$(CHz(B"	kijuhynbmlopP)
    (its-define-hangul	"gla"	"$(CH{(B"	kijuhynbmlopP)
    (its-define-hangul	"glq"	"$(CH|(B"	kijuhynbmlopP)
    (its-define-hangul	"glt"	"$(CH}(B"	kijuhynbmlopOP)
    (its-define-hangul	"gld"	"$(CH~(B"	kijuhynbmlopOP))

  (dolist (ascii '(("1" . "$(C#1(B")  ("2" . "$(C#2(B")  ("3" . "$(C#3(B")  ("4" . "$(C#4(B")
		   ("5" . "$(C#5(B")  ("6" . "$(C#6(B")  ("7" . "$(C#7(B")  ("8" . "$(C#8(B")
		   ("9" . "$(C#9(B")  ("0" . "$(C#0(B")
		   (" " . "$(C!!(B")  ("!" . "$(C#!(B")  ("@" . "$(C#@(B")  ("#" . "$(C##(B")
		   ("$" . "$(C#$(B")  ("%" . "$(C#%(B")  ("^" . "$(C#^(B")  ("&" . "$(C#&(B")
		   ("*" . "$(C#*(B")  ("(" . "$(C#((B")  (")" . "$(C#)(B")
		   ("-" . "$(C#-(B")  ("=" . "$(C#=(B")  ("`" . "$(C#`(B")  ("\\" . "$(C#\(B")
		   ("|" . "$(C#|(B")  ("_" . "$(C#_(B")  ("+" . "$(C#+(B")  ("~" . "$(C#~(B")
		   ("[" . "$(C!8(B")  ("]" . "$(C!9(B")  ("{" . "$(C#{(B")  ("}" . "$(C#}(B")
		   (":" . "$(C#:(B")  (";" . "$(C#;(B")  ("\"" . "$(C#"(B") ("'" . "$(C#'(B")
		   ("<" . "$(C#<(B")  (">" . "$(C#>(B")  ("?" . "$(C#?(B")  ("/" . "$(C#/(B")
		   ("," . "$(C#,(B")  ("." . "$(C#.(B")
		   ("a" . "$(C#a(B")  ("b" . "$(C#b(B")  ("c" . "$(C#c(B")  ("d" . "$(C#d(B")
		   ("e" . "$(C#e(B")  ("f" . "$(C#f(B")  ("g" . "$(C#g(B")  ("h" . "$(C#h(B")
		   ("i" . "$(C#i(B")  ("j" . "$(C#j(B")  ("k" . "$(C#k(B")  ("l" . "$(C#l(B")
		   ("m" . "$(C#m(B")  ("n" . "$(C#n(B")  ("o" . "$(C#o(B")  ("p" . "$(C#p(B")
		   ("q" . "$(C#q(B")  ("r" . "$(C#r(B")  ("s" . "$(C#s(B")  ("t" . "$(C#t(B")
		   ("u" . "$(C#u(B")  ("v" . "$(C#v(B")  ("w" . "$(C#w(B")  ("x" . "$(C#x(B")
		   ("y" . "$(C#y(B")  ("z" . "$(C#z(B")
		   ("A" . "$(C#A(B")  ("B" . "$(C#B(B")  ("C" . "$(C#C(B")  ("D" . "$(C#D(B")
		   ("E" . "$(C#E(B")  ("F" . "$(C#F(B")  ("G" . "$(C#G(B")  ("H" . "$(C#H(B")
		   ("I" . "$(C#I(B")  ("J" . "$(C#J(B")  ("K" . "$(C#K(B")  ("L" . "$(C#L(B")
		   ("M" . "$(C#M(B")  ("N" . "$(C#N(B")  ("O" . "$(C#O(B")  ("P" . "$(C#P(B")
		   ("Q" . "$(C#Q(B")  ("R" . "$(C#R(B")  ("S" . "$(C#S(B")  ("T" . "$(C#T(B")
		   ("U" . "$(C#U(B")  ("V" . "$(C#V(B")  ("W" . "$(C#W(B")  ("X" . "$(C#X(B")
		   ("Y" . "$(C#Y(B")  ("Z" . "$(C#Z(B")))
    (let ((in (car ascii)) (out (cdr ascii)))
      (its-defrule (concat its-pankak-escape in) in)
      (its-defrule (concat its-jeonkak-escape in) out))))

(define-its-state-machine-append its-hangul-map
  (its-defrule "[" its-hangul-open-bracket)
  (its-defrule "]" its-hangul-close-bracket)
  (its-defrule "." its-hangul-period)
  (its-defrule "," its-hangul-comma)

  (if its-hangul-enable-jeonkak-alphabet
      (progn
	(its-defrule   "1"   "$(C#1(B")  (its-defrule   "2"   "$(C#2(B")
	(its-defrule   "3"   "$(C#3(B")  (its-defrule   "4"   "$(C#4(B")
	(its-defrule   "5"   "$(C#5(B")  (its-defrule   "6"   "$(C#6(B")
	(its-defrule   "7"   "$(C#7(B")  (its-defrule   "8"   "$(C#8(B")
	(its-defrule   "9"   "$(C#9(B")  (its-defrule   "0"   "$(C#0(B")
	(its-defrule   "!"   "$(C#!(B")  (its-defrule   "@"   "$(C#@(B")
	(its-defrule   "#"   "$(C##(B")  (its-defrule   "$"   "$(C#$(B")
	(its-defrule   "%"   "$(C#%(B")  (its-defrule   "^"   "$(C#^(B")
	(its-defrule   "&"   "$(C#&(B")  (its-defrule   "*"   "$(C#*(B")
	(its-defrule   "("   "$(C#((B")  (its-defrule   ")"   "$(C#)(B")
	(its-defrule   "-"   "$(C#-(B")  (its-defrule   "~"   "$(C#~(B")
	(its-defrule   "="   "$(C#=(B")  (its-defrule   "`"   "$(C#`(B")
	(its-defrule   "\\"  "$(C#\(B")  (its-defrule   "|"   "$(C#|(B")
	(its-defrule   "_"   "$(C#_(B")  (its-defrule   "+"   "$(C#+(B")
	(its-defrule   "{"   "$(C#{(B")  (its-defrule   "}"   "$(C#}(B")
	(its-defrule   ":"   "$(C#:(B")  (its-defrule   ";"   "$(C#;(B")
	(its-defrule   "\""  "$(C#"(B")  (its-defrule   "'"   "$(C#'(B")
	(its-defrule   "<"   "$(C#<(B")  (its-defrule   ">"   "$(C#>(B")
	(its-defrule   "?"   "$(C#?(B")  (its-defrule   "/"   "$(C#/(B"))
    (progn
      (its-defrule   "1"   "1")  (its-defrule   "2"   "2")
      (its-defrule   "3"   "3")  (its-defrule   "4"   "4")
      (its-defrule   "5"   "5")  (its-defrule   "6"   "6")
      (its-defrule   "7"   "7")  (its-defrule   "8"   "8")
      (its-defrule   "9"   "9")  (its-defrule   "0"   "0")
      (its-defrule   "!"   "!")  (its-defrule   "@"   "@")
      (its-defrule   "#"   "#")  (its-defrule   "$"   "$")
      (its-defrule   "%"   "%")  (its-defrule   "^"   "^")
      (its-defrule   "&"   "&")  (its-defrule   "*"   "*")
      (its-defrule   "("   "(")  (its-defrule   ")"   ")")
      (its-defrule   "-"   "-")  (its-defrule   "~"   "~")
      (its-defrule   "="   "=")  (its-defrule   "`"   "`")
      (its-defrule   "\\"  "\\") (its-defrule   "|"   "|")
      (its-defrule   "_"   "_")  (its-defrule   "+"   "+")
      (its-defrule   "{"   "{")  (its-defrule   "}"   "}")
      (its-defrule   ":"   ":")  (its-defrule   ";"   ";")
      (its-defrule   "\""  "\"") (its-defrule   "'"   "'")
      (its-defrule   "<"   "<")  (its-defrule   ">"   ">")
      (its-defrule   "?"   "?")  (its-defrule   "/"   "/"))))

(provide 'its/hangul)
